from datetime import timedelta
from enum import Enum, EnumMeta
from typing import Dict, List, NamedTuple, Optional, Union

from kloppy.domain import (
    BallState,
    BodyPart,
    BodyPartQualifier,
    CardQualifier,
    CardType,
    CarryResult,
    CounterAttackQualifier,
    DuelQualifier,
    DuelResult,
    DuelType,
    Event,
    EventFactory,
    ExpectedGoals,
    FormationType,
    GoalkeeperActionType,
    GoalkeeperQualifier,
    InterceptionResult,
    PassQualifier,
    PassResult,
    PassType,
    PositionType,
    PostShotExpectedGoals,
    SetPieceQualifier,
    SetPieceType,
    ShotResult,
    TakeOnResult,
)
from kloppy.domain.models.event import UnderPressureQualifier
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.event.statsbomb.helpers import (
    get_period_by_id,
    get_team_by_id,
    parse_coordinates,
    parse_obv_values,
    parse_str_ts,
)


class Version(NamedTuple):
    """StatsBomb data version.

    StatsBomb maintains two levels of accuracy for coordinates in their data.
    This level can differ between Shot events and all other events.


    Attributes:
        shot_fidelity_version: The level of accuracy for Shot events and
            events paired with Shot events (these are Goal Keeper events, Block
            events, Duel events and Freeze Frame objects).
            - In V1, the x-coordinates, y-coordinates and, where appropriate,
              z-coordinates are detailed to a yard with the exception of the
              y-coordinates and z-coordinates in the goal frame of the end
              locations of Shot events.
            - In V2, the x-coordinates, y-coordinates and, where appropriate,
              z-coordinates are detailed to a tenth of a yard.
        xy_fidelity_version: The level of accuracy for all other events.
            - In V1, the x-coordinates and y-coordinates of all the locations
              associated with events are detailed to a yard.
            - In V2, the x-coordinates and y-coordinates of all the locations
              associated with events are detailed to a tenth of a yard.
    """

    shot_fidelity_version: int
    xy_fidelity_version: int


FORMATIONS = {
    3142: FormationType.THREE_ONE_FOUR_TWO,
    312112: FormationType.THREE_ONE_TWO_ONE_ONE_TWO,
    31222: FormationType.THREE_ONE_TWO_TWO_TWO,
    32122: FormationType.THREE_TWO_ONE_TWO_TWO,
    32212: FormationType.THREE_TWO_TWO_ONE_TWO,
    32221: FormationType.THREE_TWO_TWO_TWO_ONE,
    3232: FormationType.THREE_TWO_THREE_TWO,
    3322: FormationType.THREE_THREE_TWO_TWO,
    3412: FormationType.THREE_FOUR_ONE_TWO,
    3421: FormationType.THREE_FOUR_TWO_ONE,
    343: FormationType.THREE_FOUR_THREE,
    3511: FormationType.THREE_FIVE_ONE_ONE,
    352: FormationType.THREE_FIVE_TWO,
    41131: FormationType.FOUR_ONE_ONE_THREE_ONE,
    41212: FormationType.FOUR_ONE_TWO_ONE_TWO,
    41221: FormationType.FOUR_ONE_TWO_TWO_ONE,
    4132: FormationType.FOUR_ONE_THREE_TWO,
    4141: FormationType.FOUR_ONE_FOUR_ONE,
    42121: FormationType.FOUR_TWO_ONE_TWO_ONE,
    4213: FormationType.FOUR_TWO_ONE_THREE,
    42211: FormationType.FOUR_TWO_TWO_ONE_ONE,
    4222: FormationType.FOUR_TWO_TWO_TWO,
    4231: FormationType.FOUR_TWO_THREE_ONE,
    4312: FormationType.FOUR_THREE_ONE_TWO,
    4321: FormationType.FOUR_THREE_TWO_ONE,
    433: FormationType.FOUR_THREE_THREE,
    4411: FormationType.FOUR_FOUR_ONE_ONE,
    442: FormationType.FOUR_FOUR_TWO,
    451: FormationType.FOUR_FIVE_ONE,
    5221: FormationType.FIVE_TWO_TWO_ONE,
    532: FormationType.FIVE_THREE_TWO,
    541: FormationType.FIVE_FOUR_ONE,
}

position_types_mapping: Dict[int, PositionType] = {
    1: PositionType.Goalkeeper,  # Provider: Goalkeeper
    2: PositionType.RightBack,  # Provider: Right Back
    3: PositionType.RightCenterBack,  # Provider: Right Center Back
    4: PositionType.CenterBack,  # Provider: Center Back
    5: PositionType.LeftCenterBack,  # Provider: Left Center Back
    6: PositionType.LeftBack,  # Provider: Left Back
    7: PositionType.RightBack,  # Provider: Right Wing Back (mapped to Right Back)
    8: PositionType.LeftBack,  # Provider: Left Wing Back (mapped to Left Back)
    9: PositionType.RightDefensiveMidfield,  # Provider: Right Defensive Midfield
    10: PositionType.CenterDefensiveMidfield,  # Provider: Center Defensive Midfield
    11: PositionType.LeftDefensiveMidfield,  # Provider: Left Defensive Midfield
    12: PositionType.RightMidfield,  # Provider: Right Midfield
    13: PositionType.RightCentralMidfield,  # Provider: Right Center Midfield
    14: PositionType.CenterMidfield,  # Provider: Center Midfield
    15: PositionType.LeftCentralMidfield,  # Provider: Left Center Midfield
    16: PositionType.LeftMidfield,  # Provider: Left Midfield
    17: PositionType.RightWing,  # Provider: Right Wing
    18: PositionType.RightAttackingMidfield,  # Provider: Right Attacking Midfield
    19: PositionType.CenterAttackingMidfield,  # Provider: Center Attacking Midfield
    20: PositionType.LeftAttackingMidfield,  # Provider: Left Attacking Midfield
    21: PositionType.LeftWing,  # Provider: Left Wing
    22: PositionType.RightForward,  # Provider: Right Center Forward (mapped to Right Forward)
    23: PositionType.Striker,  # Provider: Striker
    24: PositionType.LeftForward,  # Provider: Left Center Forward (mapped to Left Forward)
    25: PositionType.Attacker,  # Provider: Secondary Striker (mapped to Attacker)
}


class TypesEnumMeta(EnumMeta):
    def __call__(cls, value, *args, **kw):
        if isinstance(value, dict):
            if value["id"] not in cls._value2member_map_:
                raise DeserializationError(
                    "Unknown StatsBomb {}: {}/{}".format(
                        (
                            cls.__qualname__.replace("_", " ")
                            .replace(".", " ")
                            .title()
                        ),
                        value["id"],
                        value["name"],
                    )
                )
            value = cls(value["id"])
        elif value not in cls._value2member_map_:
            raise DeserializationError(
                "Unknown StatsBomb {}: {}".format(
                    (
                        cls.__qualname__.replace("_", " ")
                        .replace(".", " ")
                        .title()
                    ),
                    value,
                )
            )
        return super().__call__(value, *args, **kw)


class EVENT_TYPE(Enum, metaclass=TypesEnumMeta):
    """The list of event types that compose all of StatsBomb data."""

    FIFTY_FIFTY = 33
    BAD_BEHAVIOUR = 24
    BALL_RECEIPT = 42
    BALL_RECOVERY = 2
    BLOCK = 6
    CAMERA_ON = 5
    CAMERA_OFF = 29
    CARRY = 43
    CLEARANCE = 9
    DISPOSSESSED = 3
    DRIBBLE = 14
    DRIBBLED_PAST = 39
    DUEL = 4
    ERROR = 37
    FOUL_COMMITTED = 22
    FOUL_WON = 21
    GOALKEEPER = 23
    HALF_END = 34
    HALF_START = 18
    INJURY_STOPPAGE = 40
    INTERCEPTION = 10
    MISCONTROL = 38
    OFFSIDE = 8
    OWN_GOAL_AGAINST = 20
    OWN_GOAL_FOR = 25
    PASS = 30
    PLAYER_OFF = 27
    PLAYER_ON = 26
    PRESSURE = 17
    REFEREE_BALL_DROP = 41
    SHIELD = 28
    SHOT = 16
    STARTING_XI = 35
    SUBSTITUTION = 19
    TACTICAL_SHIFT = 36


class BODYPART(Enum, metaclass=TypesEnumMeta):
    """The list of body parts used in StatsBomb data."""

    BOTH_HANDS = 35
    CHEST = 36
    HEAD = 37
    LEFT_FOOT = 38
    LEFT_HAND = 39
    RIGHT_FOOT = 40
    RIGHT_HAND = 41
    DROP_KICK = 68
    KEEPER_ARM = 69
    OTHER = 70
    NO_TOUCH = 106


class PLAY_PATTERN(Enum, metaclass=TypesEnumMeta):
    """The list of play patterns used in StatsBomb data."""

    REGULAR_PLAY = 1
    FROM_CORNER = 2
    FROM_FREE_KICK = 3
    FROM_THROW_IN = 4
    OTHER = 5
    FROM_COUNTER = 6
    FROM_GOAL_KICK = 7
    FROM_KEEPER = 8
    FROM_KICK_OFF = 9


class EVENT:
    """Base class for StatsBomb events.

    This class is used to deserialize StatsBomb events into kloppy events.
    This default implementation is used for all events that do not have a
    specific implementation. They are deserialized into a generic event.

    Args:
        raw_event: The raw JSON event.
        data_version: The version of the StatsBomb data.
    """

    def __init__(self, raw_event: Dict):
        self.raw_event = raw_event

    def set_version(self, data_version: Version):
        self.fidelity_version = data_version.xy_fidelity_version
        return self

    def set_refs(self, periods, teams, events):
        self.period = get_period_by_id(self.raw_event["period"], periods)
        self.team = get_team_by_id(self.raw_event["team"]["id"], teams)
        self.possession_team = get_team_by_id(
            self.raw_event["possession_team"]["id"], teams
        )
        self.player = (
            self.team.get_player_by_id(self.raw_event["player"]["id"])
            if "player" in self.raw_event
            else None
        )
        self.related_events = [
            events.get(event_id)
            for event_id in self.raw_event.get("related_events", [])
        ]
        return self

    def deserialize(self, event_factory: EventFactory) -> List[Event]:
        """Deserialize the event.

        Args:
            event_factory: The event factory to use to build the event.

        Returns:
            A list of kloppy events.
        """
        generic_event_kwargs = self._parse_generic_kwargs()

        # create events
        base_events = self._create_events(
            event_factory, **generic_event_kwargs
        )
        aerial_won_events = self._create_aerial_won_event(
            event_factory, **generic_event_kwargs
        )
        ball_out_events = self._create_ball_out_event(
            event_factory, **generic_event_kwargs
        )

        # add qualifiers
        for event in aerial_won_events + base_events:
            self._add_under_pressure_qualifier(event)
        for event in aerial_won_events + base_events + ball_out_events:
            self._add_play_pattern_qualifiers(event)

        # return events (note: order is important)
        return aerial_won_events + base_events + ball_out_events

    def _parse_generic_kwargs(self) -> Dict:
        game_state_value = parse_obv_values(self.raw_event)
        return {
            "period": self.period,
            "timestamp": parse_str_ts(self.raw_event["timestamp"]),
            "ball_owning_team": self.possession_team,
            "ball_state": BallState.ALIVE,
            "event_id": self.raw_event["id"],
            "team": self.team,
            "player": self.player,
            "coordinates": (
                parse_coordinates(
                    self.raw_event["location"],
                    self.fidelity_version,
                )
                if "location" in self.raw_event
                else None
            ),
            "related_event_ids": self.raw_event.get("related_events", []),
            "raw_event": self.raw_event,
            "statistics": [game_state_value] if game_state_value else [],
        }

    def _create_aerial_won_event(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        """Add possible aerial won - Applicable to multiple event types"""
        for type_name in ["shot", "clearance", "miscontrol", "pass"]:
            if (
                type_name in self.raw_event
                and "aerial_won" in self.raw_event[type_name]
            ):
                generic_event_kwargs[
                    "event_id"
                ] = f"duel-{generic_event_kwargs['event_id']}"
                duel_qualifiers = [
                    DuelQualifier(value=DuelType.LOOSE_BALL),
                    DuelQualifier(value=DuelType.AERIAL),
                ]
                duel_event = event_factory.build_duel(
                    result=DuelResult.WON,
                    qualifiers=duel_qualifiers,
                    **generic_event_kwargs,
                )
                return [duel_event]
        return []

    def _create_ball_out_event(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        if self.raw_event.get("out", False):
            generic_event_kwargs[
                "event_id"
            ] = f"out-{generic_event_kwargs['event_id']}"
            generic_event_kwargs["ball_state"] = BallState.DEAD
            ball_out_event = event_factory.build_ball_out(
                result=None,
                qualifiers=None,
                **generic_event_kwargs,
            )
            return [ball_out_event]
        return []

    def _add_play_pattern_qualifiers(self, event: Event) -> Event:
        if "play_pattern" in event.raw_event:
            pattern_id = PLAY_PATTERN(event.raw_event["play_pattern"]["id"])
            if pattern_id == PLAY_PATTERN.FROM_COUNTER:
                q = CounterAttackQualifier(value=True)
                event.qualifiers = event.qualifiers or []
                event.qualifiers.append(q)
        return event

    def _add_under_pressure_qualifier(self, event: Event) -> Event:
        if ("under_pressure" in self.raw_event) and (
            self.raw_event["under_pressure"]
        ):
            q = UnderPressureQualifier(True)
            event.qualifiers = event.qualifiers or []
            event.qualifiers.append(q)

        return event

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        generic_event = event_factory.build_generic(
            result=None,
            qualifiers=None,
            event_name=self.raw_event["type"]["name"],
            **generic_event_kwargs,
        )
        return [generic_event]


class PASS(EVENT):
    """StatsBomb 30/Pass event."""

    class TYPE(Enum, metaclass=TypesEnumMeta):
        ONE_TOUCH_INTERCEPTION = 64
        RECOVERY = 66
        CORNER_KICK = 61
        FREE_KICK = 62
        GOAL_KICK = 63
        KICK_OFF = 65
        THROW_IN = 67

    class OUTCOME(Enum, metaclass=TypesEnumMeta):
        COMPLETE = 8
        INCOMPLETE = 9
        INJURY_CLEARANCE = 74
        OUT = 75
        OFFSIDE = 76
        UNKNOWN = 77

    class HEIGHT(Enum, metaclass=TypesEnumMeta):
        GROUND = 1
        LOW = 2
        HIGH = 3

    class TECHNIQUE(Enum, metaclass=TypesEnumMeta):
        THROUGH_BALL = 108
        INSWINGING = 104
        OUTSWINGING = 105
        STRAIGHT = 107

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        team = generic_event_kwargs["team"]
        timestamp = generic_event_kwargs["timestamp"]
        pass_dict = self.raw_event["pass"]

        result = None
        receiver_player = None
        receiver_coordinates = parse_coordinates(
            pass_dict["end_location"],
            self.fidelity_version,
        )
        receive_timestamp = timestamp + timedelta(
            seconds=self.raw_event.get("duration", 0.0)
        )

        if "outcome" in pass_dict:
            outcome_id = pass_dict["outcome"]["id"]
            outcome_mapping = {
                PASS.OUTCOME.OUT: PassResult.OUT,
                PASS.OUTCOME.INCOMPLETE: PassResult.INCOMPLETE,
                PASS.OUTCOME.OFFSIDE: PassResult.OFFSIDE,
                PASS.OUTCOME.INJURY_CLEARANCE: PassResult.OUT,
                PASS.OUTCOME.UNKNOWN: None,
            }
            result = outcome_mapping.get(PASS.OUTCOME(outcome_id))
        else:
            result = PassResult.COMPLETE
            receiver_player = team.get_player_by_id(
                pass_dict["recipient"]["id"]
            )

        qualifiers = (
            _get_pass_qualifiers(pass_dict)
            + _get_set_piece_qualifiers(EVENT_TYPE.PASS, pass_dict)
            + _get_body_part_qualifiers(pass_dict)
        )

        pass_event = event_factory.build_pass(
            result=result,
            receive_timestamp=receive_timestamp,
            receiver_coordinates=receiver_coordinates,
            receiver_player=receiver_player,
            qualifiers=qualifiers,
            **generic_event_kwargs,
        )

        # if pass is an interception, insert interception prior to pass event
        if "type" in pass_dict:
            generic_event_kwargs[
                "event_id"
            ] = f"interception-{generic_event_kwargs['event_id']}"
            type_id = PASS.TYPE(pass_dict["type"]["id"])
            if type_id == PASS.TYPE.ONE_TOUCH_INTERCEPTION:
                interception_event = event_factory.build_interception(
                    **generic_event_kwargs,
                    result=InterceptionResult.SUCCESS,
                    qualifiers=None,
                )
                return [interception_event, pass_event]

        return [pass_event]

    def _create_ball_out_event(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        pass_dict = self.raw_event["pass"]
        if (
            self.raw_event.get("out", False)
            or "outcome" in pass_dict
            and PASS.OUTCOME(pass_dict["outcome"]) == PASS.OUTCOME.OUT
        ):
            # If there is a related (failed) ball receipt event recorded for
            # the pass, we create the ball out event from that event.
            if any(
                isinstance(related_event, BALL_RECEIPT)
                for related_event in self.related_events
            ):
                return []
            generic_event_kwargs[
                "event_id"
            ] = f"out-{generic_event_kwargs['event_id']}"
            generic_event_kwargs["ball_state"] = BallState.DEAD
            generic_event_kwargs["coordinates"] = parse_coordinates(
                pass_dict["end_location"],
                self.fidelity_version,
            )

            ball_out_event = event_factory.build_ball_out(
                result=None,
                qualifiers=None,
                **generic_event_kwargs,
            )
            return [ball_out_event]
        return []


class BALL_RECEIPT(EVENT):
    """StatsBomb 42/Ball Receipt* event."""

    def _create_ball_out_event(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        for related_event in self.related_events:
            if isinstance(related_event, PASS):
                pass_dict = related_event.raw_event.get("pass", {})
                if (
                    related_event.raw_event.get("out", False)
                    or "outcome" in pass_dict
                    and PASS.OUTCOME(pass_dict["outcome"]) == PASS.OUTCOME.OUT
                ):
                    generic_event_kwargs[
                        "event_id"
                    ] = f"out-{generic_event_kwargs['event_id']}"
                    generic_event_kwargs["ball_state"] = BallState.DEAD
                    generic_event_kwargs["coordinates"] = parse_coordinates(
                        pass_dict["end_location"],
                        self.fidelity_version,
                    )
                    generic_event_kwargs["raw_event"] = related_event.raw_event

                    ball_out_event = event_factory.build_ball_out(
                        result=None,
                        qualifiers=None,
                        **generic_event_kwargs,
                    )
                    return [ball_out_event]
                return []
        return []


class SHOT(EVENT):
    """StatsBomb 16/Shot event."""

    class TYPE(Enum, metaclass=TypesEnumMeta):
        OPEN_PLAY = 87
        FREE_KICK = 62
        KICK_OFF = 65
        CORNER_KICK = 61
        PENALTY = 88

    class OUTCOME(Enum, metaclass=TypesEnumMeta):
        BLOCKED = 96
        GOAL = 97
        OFF_TARGET = 98
        POST = 99
        SAVED = 100
        OFF_WAYWARD = 101
        SAVED_OFF_TARGET = 115
        SAVED_TO_POST = 116

    def set_version(self, data_version: Version):
        self.fidelity_version = data_version.shot_fidelity_version
        return self

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        shot_dict = self.raw_event["shot"]

        outcome_id = SHOT.OUTCOME(shot_dict["outcome"]["id"])

        outcome_mapping = {
            SHOT.OUTCOME.OFF_TARGET: ShotResult.OFF_TARGET,
            SHOT.OUTCOME.SAVED: ShotResult.SAVED,
            SHOT.OUTCOME.SAVED_OFF_TARGET: ShotResult.SAVED,
            SHOT.OUTCOME.SAVED_TO_POST: ShotResult.SAVED,
            SHOT.OUTCOME.POST: ShotResult.POST,
            SHOT.OUTCOME.OFF_WAYWARD: ShotResult.OFF_TARGET,
            SHOT.OUTCOME.BLOCKED: ShotResult.BLOCKED,
            SHOT.OUTCOME.GOAL: ShotResult.GOAL,
        }

        result = outcome_mapping.get(outcome_id)

        if result is None:
            raise DeserializationError(f"Unknown shot outcome: {outcome_id}")

        qualifiers = _get_set_piece_qualifiers(
            EVENT_TYPE.SHOT, shot_dict
        ) + _get_body_part_qualifiers(shot_dict)

        for statistic_cls, prop_name in {
            ExpectedGoals: "statsbomb_xg",
            PostShotExpectedGoals: "shot_execution_xg",
        }.items():
            value = shot_dict.get(prop_name, None)
            if value is not None:
                generic_event_kwargs["statistics"].append(
                    statistic_cls(value=value)
                )

        shot_event = event_factory.build_shot(
            result=result,
            qualifiers=qualifiers,
            result_coordinates=parse_coordinates(
                shot_dict["end_location"],
                self.fidelity_version,
            ),
            **generic_event_kwargs,
        )

        return [shot_event]

    def _create_ball_out_event(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        shot_dict = self.raw_event["shot"]
        if (
            self.raw_event.get("out", False)
            or "outcome" in shot_dict
            and SHOT.OUTCOME(shot_dict["outcome"]) == SHOT.OUTCOME.OFF_TARGET
        ):
            # If there is a related goalkeeper event recorded for
            # the shot, we create the ball out event from that event.
            if any(
                isinstance(related_event, GOALKEEPER)
                for related_event in self.related_events
            ):
                return []
            generic_event_kwargs[
                "event_id"
            ] = f"out-{generic_event_kwargs['event_id']}"
            generic_event_kwargs["ball_state"] = BallState.DEAD
            generic_event_kwargs["coordinates"] = parse_coordinates(
                shot_dict["end_location"],
                self.fidelity_version,
            )

            ball_out_event = event_factory.build_ball_out(
                result=None,
                qualifiers=None,
                **generic_event_kwargs,
            )
            return [ball_out_event]
        return []


class INTERCEPTION(EVENT):
    """StatsBomb 10/Interception event."""

    class OUTCOME(Enum, metaclass=TypesEnumMeta):
        LOST = 1
        WON = 4
        LOST_IN_PLAY = 13
        LOST_OUT = 14
        SUCCESS = 15
        SUCCESS_IN_PLAY = 16
        SUCCESS_OUT = 17

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        interception_dict = self.raw_event.get("interception", {})

        outcome = interception_dict.get("outcome", {})
        outcome_id = INTERCEPTION.OUTCOME(outcome)
        if outcome_id in [
            INTERCEPTION.OUTCOME.LOST_OUT,
            INTERCEPTION.OUTCOME.SUCCESS_OUT,
        ]:
            result = InterceptionResult.OUT
        elif outcome_id in [
            INTERCEPTION.OUTCOME.WON,
            INTERCEPTION.OUTCOME.SUCCESS,
            INTERCEPTION.OUTCOME.SUCCESS_IN_PLAY,
        ]:
            result = InterceptionResult.SUCCESS
        elif outcome_id in [
            INTERCEPTION.OUTCOME.LOST,
            INTERCEPTION.OUTCOME.LOST_IN_PLAY,
        ]:
            result = InterceptionResult.LOST
        else:
            raise DeserializationError(
                f"Unknown interception outcome: {outcome.get('name')}({outcome_id})"
            )

        interception_event = event_factory.build_interception(
            result=result,
            qualifiers=None,
            **generic_event_kwargs,
        )

        return [interception_event]

    def _create_ball_out_event(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        interception_dict = self.raw_event.get("interception", {})
        if (
            self.raw_event.get("out", False)
            or "outcome" in interception_dict
            and INTERCEPTION.OUTCOME(interception_dict["outcome"])
            in [
                INTERCEPTION.OUTCOME.LOST_OUT,
                INTERCEPTION.OUTCOME.SUCCESS_OUT,
            ]
        ):
            generic_event_kwargs[
                "event_id"
            ] = f"out-{generic_event_kwargs['event_id']}"
            generic_event_kwargs["ball_state"] = BallState.DEAD
            ball_out_event = event_factory.build_ball_out(
                result=None,
                qualifiers=None,
                **generic_event_kwargs,
            )
            return [ball_out_event]
        return []


class OWN_GOAL_AGAINST(EVENT):
    """StatsBomb 20/Own goal against event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        shot_event = event_factory.build_shot(
            result=ShotResult.OWN_GOAL,
            qualifiers=None,
            **generic_event_kwargs,
        )
        return [shot_event]


class OWN_GOAL_FOR(EVENT):
    """StatsBomb 25/Own goal for event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        return []


class CLEARANCE(EVENT):
    """StatsBomb 9/Clearance event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        clearance_dict = self.raw_event.get("clearance", {})
        # Old versions of the data (< v1.1) don't define extra attributes for clearances
        qualifiers = _get_body_part_qualifiers(clearance_dict)

        clearance_event = event_factory.build_clearance(
            result=None,
            qualifiers=qualifiers,
            **generic_event_kwargs,
        )

        return [clearance_event]


class MISCONTROL(EVENT):
    """StatsBomb 38/Miscontrol event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        miscontrol_event = event_factory.build_miscontrol(
            result=None,
            qualifiers=None,
            **generic_event_kwargs,
        )

        return [miscontrol_event]


class DRIBBLE(EVENT):
    """StatsBomb 14/Dribble event."""

    class OUTCOME(Enum, metaclass=TypesEnumMeta):
        COMPLETE = 8
        INCOMPLETE = 9

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        dribble_dict = self.raw_event.get("dribble", {})
        result_mapping = {
            DRIBBLE.OUTCOME.INCOMPLETE: TakeOnResult.INCOMPLETE,
            DRIBBLE.OUTCOME.COMPLETE: TakeOnResult.COMPLETE,
        }

        result = TakeOnResult.COMPLETE
        if "outcome" in dribble_dict:
            outcome_id = DRIBBLE.OUTCOME(dribble_dict["outcome"])
            result = result_mapping[outcome_id]

        if result == TakeOnResult.INCOMPLETE:
            for related_event in self.related_events:
                if isinstance(related_event, DUEL):
                    duel_dict = related_event.raw_event.get("duel", {})
                    if "outcome" in duel_dict and DUEL.OUTCOME(
                        duel_dict["outcome"]
                    ) in [DUEL.OUTCOME.SUCCESS_OUT, DUEL.OUTCOME.LOST_OUT]:
                        result = TakeOnResult.OUT
                        break

        take_on_event = event_factory.build_take_on(
            qualifiers=None,
            result=result,
            **generic_event_kwargs,
        )
        return [take_on_event]


class CARRY(EVENT):
    """StatsBomb 43/Carry event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        timestamp = generic_event_kwargs["timestamp"]
        carry_dict = self.raw_event["carry"]
        carry_event = event_factory.build_carry(
            qualifiers=None,
            end_timestamp=timestamp
            + timedelta(seconds=self.raw_event.get("duration", 0)),
            result=CarryResult.COMPLETE,
            end_coordinates=parse_coordinates(
                carry_dict["end_location"],
                self.fidelity_version,
            ),
            **generic_event_kwargs,
        )
        return [carry_event]


class DUEL(EVENT):
    """StatsBomb 4/Duel event."""

    class TYPE(Enum, metaclass=TypesEnumMeta):
        AERIAL_LOST = 10
        TACKLE = 11

    class OUTCOME(Enum, metaclass=TypesEnumMeta):
        # Recorded when DUEL.TYPE is DUEL_TACKLE
        WON = 4
        LOST_IN_PLAY = 13
        LOST_OUT = 14
        SUCCESS = 15
        SUCCESS_IN_PLAY = 16
        SUCCESS_OUT = 17

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        duel_dict = self.raw_event.get("duel", {})

        # Get duel qualifiers
        type_id = DUEL.TYPE(duel_dict["type"])
        duel_qualifiers = []
        if type_id == DUEL.TYPE.AERIAL_LOST:
            duel_qualifiers = [
                DuelQualifier(value=DuelType.LOOSE_BALL),
                DuelQualifier(value=DuelType.AERIAL),
            ]
        elif type_id == DUEL.TYPE.TACKLE:
            duel_qualifiers = [DuelQualifier(value=DuelType.GROUND)]

        # Get duel result
        duel_won_outcomes = [
            DUEL.OUTCOME.WON,
            DUEL.OUTCOME.SUCCESS,
            DUEL.OUTCOME.SUCCESS_IN_PLAY,
            DUEL.OUTCOME.SUCCESS_OUT,
        ]
        result = (
            DuelResult.WON
            if type_id != DUEL.TYPE.AERIAL_LOST
            and DUEL.OUTCOME(duel_dict.get("outcome", {})) in duel_won_outcomes
            else DuelResult.LOST
        )

        duel_event = event_factory.build_duel(
            result=result,
            qualifiers=duel_qualifiers,
            **generic_event_kwargs,
        )
        return [duel_event]

    def _create_ball_out_event(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        duel_dict = self.raw_event.get("duel", {})
        if (
            self.raw_event.get("out", False)
            or "outcome" in duel_dict
            and DUEL.OUTCOME(duel_dict["outcome"])
            in [DUEL.OUTCOME.LOST_OUT, DUEL.OUTCOME.SUCCESS_OUT]
        ):
            generic_event_kwargs[
                "event_id"
            ] = f"out-{generic_event_kwargs['event_id']}"
            generic_event_kwargs["ball_state"] = BallState.DEAD
            ball_out_event = event_factory.build_ball_out(
                result=None,
                qualifiers=None,
                **generic_event_kwargs,
            )
            return [ball_out_event]
        return []


class FIFTY_FIFTY(EVENT):
    """StatsBomb 33/Fifty-Fifty event."""

    class OUTCOME(Enum, metaclass=TypesEnumMeta):
        WON = 4
        LOST = 1
        SUCCESS_TO_TEAM = 3
        SUCCESS_TO_OPPONENT = 2

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        duel_dict = self.raw_event.get("50_50", {})

        # Get duel qualifiers
        duel_qualifiers = [
            DuelQualifier(value=DuelType.LOOSE_BALL),
            DuelQualifier(value=DuelType.GROUND),
        ]

        # Get duel result
        duel_won_outcomes = [
            FIFTY_FIFTY.OUTCOME.WON,
            FIFTY_FIFTY.OUTCOME.SUCCESS_TO_TEAM,
        ]
        result = (
            DuelResult.WON
            if FIFTY_FIFTY.OUTCOME(duel_dict.get("outcome", {}))
            in duel_won_outcomes
            else DuelResult.LOST
        )

        duel_event = event_factory.build_duel(
            result=result,
            qualifiers=duel_qualifiers,
            **generic_event_kwargs,
        )
        return [duel_event]


class GOALKEEPER(EVENT):
    """StatsBomb 23/Goalkeeper event."""

    class TYPE(Enum, metaclass=TypesEnumMeta):
        COLLECTED = 25
        GOAL_CONCEDED = 26
        KEEPER_SWEEPER = 27
        PENALTY_CONCEDED = 28
        PENALTY_SAVED = 29
        PUNCH = 30
        PENALTY_SAVED_TO_POST = 109
        SAVE = 31
        SAVED_TO_POST = 110  # A save by the goalkeeper that hits the post
        SHOT_FACED = 32
        SHOT_SAVED = 33
        SHOT_SAVED_OFF_TARGET = 113
        SHOT_SAVED_TO_POST = (
            114  # A shot saved by the goalkeeper that hits the post
        )
        SMOTHER = 34

    class KEEPER_SWEEPER:
        class OUTCOME(Enum, metaclass=TypesEnumMeta):
            CLAIM = 47
            CLEAR = 48
            WON = 4
            SUCCESS = 15

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        goalkeeper_dict = self.raw_event["goalkeeper"]
        generic_event_kwargs = self._parse_generic_kwargs()

        # parse body part
        body_part_qualifiers = _get_body_part_qualifiers(goalkeeper_dict)
        hands_used = any(
            q.value
            in [
                BodyPart.LEFT_HAND,
                BodyPart.RIGHT_HAND,
                BodyPart.BOTH_HANDS,
            ]
            for q in body_part_qualifiers
        )
        head_or_foot_used = any(
            q.value
            in [
                BodyPart.LEFT_FOOT,
                BodyPart.RIGHT_FOOT,
                BodyPart.HEAD,
            ]
            for q in body_part_qualifiers
        )
        bodypart_missing = len(body_part_qualifiers) == 0

        # parse action type qualifiers
        save_event_types = [
            GOALKEEPER.TYPE.SHOT_SAVED,
            GOALKEEPER.TYPE.PENALTY_SAVED_TO_POST,
            GOALKEEPER.TYPE.SAVED_TO_POST,
            GOALKEEPER.TYPE.SHOT_SAVED_OFF_TARGET,
            GOALKEEPER.TYPE.SHOT_SAVED_TO_POST,
        ]
        type_id = GOALKEEPER.TYPE(goalkeeper_dict.get("type", {}).get("id"))
        outcome_id = goalkeeper_dict.get("outcome", {}).get("id")
        qualifiers = []
        if type_id in save_event_types:
            qualifiers.append(
                GoalkeeperQualifier(value=GoalkeeperActionType.SAVE)
            )
        elif type_id == GOALKEEPER.TYPE.SMOTHER:
            qualifiers.append(
                GoalkeeperQualifier(value=GoalkeeperActionType.SMOTHER)
            )
        elif type_id == GOALKEEPER.TYPE.PUNCH:
            qualifiers.append(
                GoalkeeperQualifier(value=GoalkeeperActionType.PUNCH)
            )
        elif type_id == GOALKEEPER.TYPE.COLLECTED:
            qualifiers.append(
                GoalkeeperQualifier(value=GoalkeeperActionType.CLAIM)
            )
        elif type_id == GOALKEEPER.TYPE.KEEPER_SWEEPER:
            outcome_id = GOALKEEPER.KEEPER_SWEEPER.OUTCOME(
                goalkeeper_dict.get("outcome", {}).get("id")
            )
            if outcome_id == GOALKEEPER.KEEPER_SWEEPER.OUTCOME.CLAIM:
                # a goalkeeper can only pick up the ball with his hands
                if hands_used or bodypart_missing:
                    qualifiers.append(
                        GoalkeeperQualifier(value=GoalkeeperActionType.PICK_UP)
                    )
                # otherwise it's a recovery
                else:
                    recovery = event_factory.build_recovery(
                        result=None,
                        qualifiers=body_part_qualifiers,
                        **generic_event_kwargs,
                    )
                    return [recovery]
            elif outcome_id in [
                GOALKEEPER.KEEPER_SWEEPER.OUTCOME.CLEAR,
                GOALKEEPER.KEEPER_SWEEPER.OUTCOME.SUCCESS,
            ]:
                # if the goalkeeper uses his foot or head, it's a clearance
                if head_or_foot_used:
                    clearance = event_factory.build_clearance(
                        result=None,
                        qualifiers=body_part_qualifiers,
                        **generic_event_kwargs,
                    )
                    return [clearance]
                # otherwise, it's a save
                else:
                    qualifiers.append(
                        GoalkeeperQualifier(value=GoalkeeperActionType.SAVE)
                    )

        if qualifiers:
            goalkeeper_event = event_factory.build_goalkeeper_event(
                result=None,
                qualifiers=qualifiers + body_part_qualifiers,
                **generic_event_kwargs,
            )
            return [goalkeeper_event]

        generic_event = event_factory.build_generic(
            result=None,
            qualifiers=None,
            event_name=self.raw_event["type"]["name"],
            **generic_event_kwargs,
        )
        return [generic_event]

    def _create_ball_out_event(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        goalkeeper_dict = self.raw_event["goalkeeper"]
        if (
            self.raw_event.get("out", False)
            or "outcome" in goalkeeper_dict
            and "Out" in goalkeeper_dict["outcome"]["name"]
        ):
            generic_event_kwargs[
                "event_id"
            ] = f"out-{generic_event_kwargs['event_id']}"
            generic_event_kwargs["ball_state"] = BallState.DEAD
            ball_out_event = event_factory.build_ball_out(
                result=None,
                qualifiers=None,
                **generic_event_kwargs,
            )
            return [ball_out_event]
        for related_event in self.related_events:
            if isinstance(related_event, SHOT):
                shot_dict = related_event.raw_event.get("shot", {})
                if (
                    related_event.raw_event.get("out", False)
                    or "outcome" in shot_dict
                    and SHOT.OUTCOME(shot_dict["outcome"])
                    == SHOT.OUTCOME.OFF_TARGET
                ):
                    generic_event_kwargs[
                        "event_id"
                    ] = f"out-{generic_event_kwargs['event_id']}"
                    generic_event_kwargs["ball_state"] = BallState.DEAD
                    generic_event_kwargs["coordinates"] = parse_coordinates(
                        shot_dict["end_location"],
                        self.fidelity_version,
                    )
                    generic_event_kwargs["raw_event"] = related_event.raw_event

                    ball_out_event = event_factory.build_ball_out(
                        result=None,
                        qualifiers=None,
                        **generic_event_kwargs,
                    )
                    return [ball_out_event]
                return []
        return []


class SUBSTITUTION(EVENT):
    """StatsBomb 19/Substitution event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        team = generic_event_kwargs["team"]
        substitution_dict = self.raw_event["substitution"]

        replacement_player_id = substitution_dict["replacement"]["id"]
        replacement_player = next(
            (
                player
                for player in team.players
                if player.player_id == str(replacement_player_id)
            ),
            None,
        )

        if replacement_player is None:
            raise DeserializationError(
                f"Could not find replacement player {replacement_player_id}"
            )

        substitution_event = event_factory.build_substitution(
            result=None,
            qualifiers=None,
            replacement_player=replacement_player,
            **generic_event_kwargs,
        )
        return [substitution_event]


class BAD_BEHAVIOUR(EVENT):
    """StatsBomb 24/Bad behaviour event."""

    class CARD(Enum, metaclass=TypesEnumMeta):
        FIRST_YELLOW = 7
        SECOND_YELLOW = 6
        RED = 5

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        bad_behaviour_dict = self.raw_event.get("bad_behaviour", {})
        card_type = _get_card_type(
            EVENT_TYPE.BAD_BEHAVIOUR, bad_behaviour_dict
        )
        if card_type:
            card_event = event_factory.build_card(
                result=None,
                qualifiers=None,
                card_type=card_type,
                **generic_event_kwargs,
            )
            return [card_event]

        generic_event = event_factory.build_generic(
            result=None,
            qualifiers=None,
            event_name=self.raw_event["type"]["name"],
            **generic_event_kwargs,
        )
        return [generic_event]


class FOUL_COMMITTED(EVENT):
    """StatsBomb 22/Foul committed event."""

    class CARD(Enum, metaclass=TypesEnumMeta):
        FIRST_YELLOW = 7
        SECOND_YELLOW = 6
        RED = 5

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        foul_committed_dict = self.raw_event.get("foul_committed", {})
        card_type = _get_card_type(
            EVENT_TYPE.FOUL_COMMITTED, foul_committed_dict
        )
        if card_type:
            foul_committed_event = event_factory.build_foul_committed(
                result=None,
                qualifiers=[CardQualifier(value=card_type)],
                **generic_event_kwargs,
            )
            card_event = event_factory.build_card(
                result=None,
                qualifiers=None,
                card_type=card_type,
                **generic_event_kwargs,
            )
            return [foul_committed_event, card_event]

        foul_committed_event = event_factory.build_foul_committed(
            result=None,
            qualifiers=None,
            **generic_event_kwargs,
        )
        return [foul_committed_event]


class PLAYER_ON(EVENT):
    """StatsBomb 26/Player on event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        player_on_event = event_factory.build_player_on(
            result=None,
            qualifiers=None,
            **generic_event_kwargs,
        )
        return [player_on_event]


class PLAYER_OFF(EVENT):
    """StatsBomb 27/Player off event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        player_off_event = event_factory.build_player_off(
            result=None,
            qualifiers=None,
            **generic_event_kwargs,
        )
        return [player_off_event]


class BALL_RECOVERY(EVENT):
    """StatsBomb 2/Ball recovery event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        recovery_dict = self.raw_event.get("ball_recovery", {})
        recovery_failure = recovery_dict.get("recovery_failure", False)
        if recovery_failure:
            duel_event = event_factory.build_duel(
                result=DuelResult.LOST,
                qualifiers=[
                    DuelQualifier(value=DuelType.LOOSE_BALL),
                ],
                **generic_event_kwargs,
            )
            return [duel_event]

        recovery_event = event_factory.build_recovery(
            result=None,
            qualifiers=None,
            **generic_event_kwargs,
        )
        return [recovery_event]


class PRESSURE(EVENT):
    """StatsBomb 17/Pressure event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        end_timestamp = generic_event_kwargs["timestamp"] + timedelta(
            seconds=self.raw_event.get("duration", 0.0)
        )

        pressure_event = event_factory.build_pressure_event(
            result=None,
            qualifiers=None,
            end_timestamp=end_timestamp,
            **generic_event_kwargs,
        )
        return [pressure_event]


class TACTICAL_SHIFT(EVENT):
    """StatsBomb 36/Tactical shift event."""

    def _create_events(
        self, event_factory: EventFactory, **generic_event_kwargs
    ) -> List[Event]:
        formation = FORMATIONS[self.raw_event["tactics"]["formation"]]
        player_positions = {}
        team = generic_event_kwargs["team"]
        for player in self.raw_event["tactics"]["lineup"]:
            player_positions[
                team.get_player_by_id(player["player"]["id"])
            ] = position_types_mapping[player["position"]["id"]]

        formation_change_event = event_factory.build_formation_change(
            result=None,
            qualifiers=None,
            formation_type=formation,
            player_positions=player_positions,
            **generic_event_kwargs,
        )
        return [formation_change_event]


def _get_card_type(
    event_type: EVENT_TYPE, event_dict: Dict
) -> Optional[CardType]:
    sb_to_kloppy_card_mappings = {
        FOUL_COMMITTED.CARD.FIRST_YELLOW: CardType.FIRST_YELLOW,
        FOUL_COMMITTED.CARD.SECOND_YELLOW: CardType.SECOND_YELLOW,
        FOUL_COMMITTED.CARD.RED: CardType.RED,
        BAD_BEHAVIOUR.CARD.FIRST_YELLOW: CardType.FIRST_YELLOW,
        BAD_BEHAVIOUR.CARD.SECOND_YELLOW: CardType.SECOND_YELLOW,
        BAD_BEHAVIOUR.CARD.RED: CardType.RED,
    }
    if "card" in event_dict:
        if event_type == EVENT_TYPE.FOUL_COMMITTED:
            card_id = FOUL_COMMITTED.CARD(event_dict["card"])
        elif event_type == EVENT_TYPE.BAD_BEHAVIOUR:
            card_id = BAD_BEHAVIOUR.CARD(event_dict["card"])
        return sb_to_kloppy_card_mappings[card_id]
    return None


def _get_body_part_qualifiers(
    event_dict: Dict,
) -> List[BodyPartQualifier]:
    sb_to_kloppy_body_part_mapping = {
        BODYPART.BOTH_HANDS: BodyPart.BOTH_HANDS,
        BODYPART.CHEST: BodyPart.CHEST,
        BODYPART.HEAD: BodyPart.HEAD,
        BODYPART.LEFT_FOOT: BodyPart.LEFT_FOOT,
        BODYPART.LEFT_HAND: BodyPart.LEFT_HAND,
        BODYPART.RIGHT_FOOT: BodyPart.RIGHT_FOOT,
        BODYPART.RIGHT_HAND: BodyPart.RIGHT_HAND,
        BODYPART.DROP_KICK: BodyPart.DROP_KICK,
        BODYPART.KEEPER_ARM: BodyPart.KEEPER_ARM,
        BODYPART.OTHER: BodyPart.OTHER,
        BODYPART.NO_TOUCH: BodyPart.NO_TOUCH,
    }

    if "body_part" in event_dict:
        body_part_id = BODYPART(event_dict["body_part"])
        if body_part_id in sb_to_kloppy_body_part_mapping:
            body_part = sb_to_kloppy_body_part_mapping[body_part_id]
            return [BodyPartQualifier(value=body_part)]

    return []


def _get_pass_qualifiers(pass_dict: Dict) -> List[PassQualifier]:
    qualifiers = []

    add_qualifier = lambda value: qualifiers.append(PassQualifier(value=value))

    if "cross" in pass_dict:
        add_qualifier(PassType.CROSS)
    if "technique" in pass_dict:
        technique_id = PASS.TECHNIQUE(pass_dict["technique"])
        if technique_id == PASS.TECHNIQUE.THROUGH_BALL:
            add_qualifier(PassType.THROUGH_BALL)
    if "switch" in pass_dict:
        add_qualifier(PassType.SWITCH_OF_PLAY)
    if "height" in pass_dict:
        height_id = PASS.HEIGHT(pass_dict["height"])
        if height_id == PASS.HEIGHT.HIGH:
            add_qualifier(PassType.HIGH_PASS)
    if "length" in pass_dict:
        pass_length = pass_dict["length"]
        if pass_length > 35:  # adopt Opta definition: 32 meters -> 35 yards
            add_qualifier(PassType.LONG_BALL)
    if "body_part" in pass_dict:
        body_part_id = BODYPART(pass_dict["body_part"])
        if body_part_id == BODYPART.HEAD:
            add_qualifier(PassType.HEAD_PASS)
        elif body_part_id == BODYPART.KEEPER_ARM:
            add_qualifier(PassType.HAND_PASS)
    if "shot_assist" in pass_dict:
        add_qualifier(PassType.SHOT_ASSIST)
    if "goal_assist" in pass_dict:
        add_qualifier(PassType.ASSIST)
        add_qualifier(PassType.SHOT_ASSIST)
    return qualifiers


def _get_set_piece_qualifiers(
    event_type: EVENT_TYPE, event_dict: Dict
) -> List[SetPieceQualifier]:
    sb_to_kloppy_set_piece_mapping = {
        PASS.TYPE.CORNER_KICK: SetPieceType.CORNER_KICK,
        SHOT.TYPE.CORNER_KICK: SetPieceType.CORNER_KICK,
        PASS.TYPE.FREE_KICK: SetPieceType.FREE_KICK,
        SHOT.TYPE.FREE_KICK: SetPieceType.FREE_KICK,
        SHOT.TYPE.PENALTY: SetPieceType.PENALTY,
        PASS.TYPE.THROW_IN: SetPieceType.THROW_IN,
        PASS.TYPE.KICK_OFF: SetPieceType.KICK_OFF,
        SHOT.TYPE.KICK_OFF: SetPieceType.KICK_OFF,
        PASS.TYPE.GOAL_KICK: SetPieceType.GOAL_KICK,
    }

    if "type" in event_dict:
        type_id = None
        if event_type == EVENT_TYPE.PASS:
            type_id = PASS.TYPE(event_dict["type"]["id"])
        elif event_type == EVENT_TYPE.SHOT:
            type_id = SHOT.TYPE(event_dict["type"]["id"])
        if type_id in sb_to_kloppy_set_piece_mapping:
            set_piece_type = sb_to_kloppy_set_piece_mapping[type_id]
            return [SetPieceQualifier(value=set_piece_type)]

    return []


def event_decoder(raw_event: Dict) -> Union[EVENT, Dict]:
    type_to_event = {
        EVENT_TYPE.PASS: PASS,
        EVENT_TYPE.BALL_RECEIPT: BALL_RECEIPT,
        EVENT_TYPE.SHOT: SHOT,
        EVENT_TYPE.INTERCEPTION: INTERCEPTION,
        EVENT_TYPE.OWN_GOAL_FOR: OWN_GOAL_FOR,
        EVENT_TYPE.OWN_GOAL_AGAINST: OWN_GOAL_AGAINST,
        EVENT_TYPE.CLEARANCE: CLEARANCE,
        EVENT_TYPE.MISCONTROL: MISCONTROL,
        EVENT_TYPE.DRIBBLE: DRIBBLE,
        EVENT_TYPE.CARRY: CARRY,
        EVENT_TYPE.DUEL: DUEL,
        EVENT_TYPE.FIFTY_FIFTY: FIFTY_FIFTY,
        EVENT_TYPE.GOALKEEPER: GOALKEEPER,
        EVENT_TYPE.SUBSTITUTION: SUBSTITUTION,
        EVENT_TYPE.BAD_BEHAVIOUR: BAD_BEHAVIOUR,
        EVENT_TYPE.FOUL_COMMITTED: FOUL_COMMITTED,
        EVENT_TYPE.PLAYER_ON: PLAYER_ON,
        EVENT_TYPE.PLAYER_OFF: PLAYER_OFF,
        EVENT_TYPE.BALL_RECOVERY: BALL_RECOVERY,
        EVENT_TYPE.PRESSURE: PRESSURE,
        EVENT_TYPE.TACTICAL_SHIFT: TACTICAL_SHIFT,
    }
    event_type = EVENT_TYPE(raw_event["type"])
    event_creator = type_to_event.get(event_type, EVENT)
    return event_creator(raw_event)



import warnings
from dataclasses import fields, replace
from typing import Optional, Union

from kloppy.domain import (
    DEFAULT_PITCH_LENGTH,
    DEFAULT_PITCH_WIDTH,
    AttackingDirection,
    CoordinateSystem,
    CustomCoordinateSystem,
    Dataset,
    DatasetFlag,
    DatasetType,
    EventDataset,
    Frame,
    Orientation,
    Period,
    PitchDimensions,
    Point,
    Point3D,
    Provider,
    ProviderCoordinateSystem,
    Team,
    TrackingDataset,
    build_coordinate_system,
)
from kloppy.domain.models.event import Event
from kloppy.domain.models.tracking import PlayerData
from kloppy.exceptions import KloppyError


class DatasetTransformer:
    def __init__(
        self,
        from_coordinate_system: Optional[CoordinateSystem] = None,
        from_pitch_dimensions: Optional[PitchDimensions] = None,
        from_orientation: Optional[Orientation] = None,
        to_coordinate_system: Optional[CoordinateSystem] = None,
        to_pitch_dimensions: Optional[PitchDimensions] = None,
        to_orientation: Optional[Orientation] = None,
    ):
        if (
            from_pitch_dimensions
            and from_coordinate_system
            or to_pitch_dimensions
            and to_coordinate_system
        ):
            raise ValueError(
                "You can't specify both a PitchDimension and CoordinateSystem transformation on the same Transformer instance"
            )

        self._from_coordinate_system = from_coordinate_system
        if from_pitch_dimensions:
            self._from_pitch_dimensions = from_pitch_dimensions
        elif from_coordinate_system:
            self._from_pitch_dimensions = (
                from_coordinate_system.pitch_dimensions
            )
        else:
            raise ValueError(
                "You must either specify the source PitchDimension or CoordinateSystem"
            )

        self._to_coordinate_system = to_coordinate_system
        if to_pitch_dimensions:
            if from_pitch_dimensions is None:
                raise ValueError(
                    "You must specify the source PitchDimension when specifying the target PitchDimension"
                )
            self._to_pitch_dimensions = to_pitch_dimensions
        elif to_coordinate_system:
            if from_coordinate_system is None:
                raise ValueError(
                    "You must specify the source CoordinateSystem when specifying the target CoordinateSystem"
                )
            self._to_pitch_dimensions = to_coordinate_system.pitch_dimensions

        self._from_orientation = from_orientation
        self._to_orientation = to_orientation
        if (
            from_orientation
            and not to_orientation
            or not from_orientation
            and to_orientation
        ):
            raise ValueError(
                "You must specify both the source and target Orientation"
            )

    @property
    def _needs_coordinate_system_change(self):
        return self._from_coordinate_system != self._to_coordinate_system

    @property
    def _needs_pitch_dimensions_change(self):
        return self._from_pitch_dimensions != self._to_pitch_dimensions

    @property
    def _needs_orientation_change(self):
        return self._from_orientation != self._to_orientation

    def change_point_dimensions(
        self, point: Union[Point, Point3D, None]
    ) -> Union[Point, Point3D, None]:
        if point is None:
            return None

        base_pitch_length = (
            self._from_pitch_dimensions.pitch_length or DEFAULT_PITCH_LENGTH
        )
        base_pitch_width = (
            self._from_pitch_dimensions.pitch_width or DEFAULT_PITCH_WIDTH
        )

        point_base = self._from_pitch_dimensions.to_metric_base(
            point, pitch_length=base_pitch_length, pitch_width=base_pitch_width
        )
        print(point_base)
        print(self._to_pitch_dimensions.from_metric_base)
        point_to = self._to_pitch_dimensions.from_metric_base(
            point=point_base,
            pitch_length=base_pitch_length,
            pitch_width=base_pitch_width,
        )

        return point_to

    def flip_point(
        self, point: Union[Point, Point3D, None]
    ) -> Union[Point, Point3D, None]:
        if not point:
            return None

        x_base = self._to_pitch_dimensions.x_dim.to_base(point.x)
        y_base = self._to_pitch_dimensions.y_dim.to_base(point.y)

        x_base = 1 - x_base
        y_base = 1 - y_base

        x = self._to_pitch_dimensions.x_dim.from_base(x_base)
        y = self._to_pitch_dimensions.y_dim.from_base(y_base)

        if isinstance(point, Point3D):
            return Point3D(x=x, y=y, z=point.z)
        else:
            return Point(x=x, y=y)

    def __needs_flip(
        self,
        period: Period,
        ball_owning_team: Optional[Team] = None,
        action_executing_team: Optional[Team] = None,
    ) -> bool:
        if (
            self._from_orientation is None
            or self._to_orientation is None
            or self._from_orientation == self._to_orientation
        ):
            flip = False
        else:
            if action_executing_team is None:
                action_executing_team = ball_owning_team

            attacking_direction_from = AttackingDirection.from_orientation(
                self._from_orientation,
                period=period,
                ball_owning_team=ball_owning_team,
                action_executing_team=action_executing_team,
            )
            attacking_direction_to = AttackingDirection.from_orientation(
                self._to_orientation,
                period=period,
                ball_owning_team=ball_owning_team,
                action_executing_team=action_executing_team,
            )
            flip = (
                attacking_direction_from != attacking_direction_to
                and attacking_direction_to != AttackingDirection.NOT_SET
            )
        return flip

    def transform_frame(self, frame: Frame) -> Frame:
        # Change coordinate system
        if self._needs_coordinate_system_change:
            frame = self.__change_frame_coordinate_system(frame)

        # Change dimensions
        elif self._needs_pitch_dimensions_change:
            frame = self.__change_frame_dimensions(frame)

        # Flip frame based on orientation
        if self._needs_orientation_change:
            if self.__needs_flip(
                period=frame.period,
                ball_owning_team=frame.ball_owning_team,
            ):
                frame = self.__flip_frame(frame)

        return frame

    def __change_frame_coordinate_system(self, frame: Frame):
        return Frame(
            # doesn't change
            timestamp=frame.timestamp,
            frame_id=frame.frame_id,
            ball_owning_team=frame.ball_owning_team,
            ball_state=frame.ball_state,
            period=frame.period,
            # changes
            ball_coordinates=self.__change_point_coordinate_system(
                frame.ball_coordinates
            ),
            ball_speed=frame.ball_speed,
            players_data={
                key: PlayerData(
                    coordinates=self.__change_point_coordinate_system(
                        player_data.coordinates
                    ),
                    distance=player_data.distance,
                    speed=player_data.speed,
                    other_data=player_data.other_data,
                )
                for key, player_data in frame.players_data.items()
            },
            other_data=frame.other_data,
            statistics=frame.statistics,
        )

    def __change_frame_dimensions(self, frame: Frame):
        return Frame(
            # doesn't change
            timestamp=frame.timestamp,
            frame_id=frame.frame_id,
            ball_owning_team=frame.ball_owning_team,
            ball_state=frame.ball_state,
            period=frame.period,
            # changes
            ball_coordinates=self.change_point_dimensions(
                frame.ball_coordinates
            ),
            players_data={
                key: PlayerData(
                    coordinates=self.change_point_dimensions(
                        player_data.coordinates
                    ),
                    distance=player_data.distance,
                    speed=player_data.speed,
                    other_data=player_data.other_data,
                )
                for key, player_data in frame.players_data.items()
            },
            other_data=frame.other_data,
            statistics=frame.statistics,
        )

    def __change_point_coordinate_system(
        self, point: Union[Point, Point3D, None]
    ) -> Union[Point, Point3D, None]:
        if not point:
            return None

        base_pitch_length = (
            self._from_pitch_dimensions.pitch_length or DEFAULT_PITCH_LENGTH
        )
        base_pitch_width = (
            self._from_pitch_dimensions.pitch_width or DEFAULT_PITCH_WIDTH
        )

        point_base = self._from_pitch_dimensions.to_metric_base(
            point, pitch_length=base_pitch_length, pitch_width=base_pitch_width
        )

        if (
            self._from_coordinate_system.vertical_orientation
            != self._to_coordinate_system.vertical_orientation
        ):
            point_base = replace(
                point_base,
                y=base_pitch_width - point_base.y,
            )

        point_to = self._to_pitch_dimensions.from_metric_base(
            point_base,
            pitch_length=base_pitch_length,
            pitch_width=base_pitch_width,
        )

        return point_to

    def __flip_frame(self, frame: Frame):
        players_data = {}
        for player, data in frame.players_data.items():
            players_data[player] = PlayerData(
                coordinates=self.flip_point(data.coordinates),
                distance=data.distance,
                speed=data.speed,
                other_data=data.other_data,
            )

        return Frame(
            # doesn't change
            timestamp=frame.timestamp,
            frame_id=frame.frame_id,
            ball_owning_team=frame.ball_owning_team,
            ball_state=frame.ball_state,
            period=frame.period,
            # changes
            ball_coordinates=self.flip_point(frame.ball_coordinates),
            players_data=players_data,
            other_data=frame.other_data,
            statistics=frame.statistics,
        )

    def transform_event(self, event: Event) -> Event:
        # Change coordinate system
        if self._needs_coordinate_system_change:
            event = self.__change_event_coordinate_system(event)

        # Change dimensions
        elif self._needs_pitch_dimensions_change:
            event = self.__change_event_dimensions(event)

        # Flip event based on orientation
        if self._needs_orientation_change:
            if self.__needs_flip(
                period=event.period,
                ball_owning_team=event.ball_owning_team,
                action_executing_team=event.team,
            ):
                event = self.__flip_event(event)

            if event.freeze_frame:
                event.freeze_frame = self.transform_frame(event.freeze_frame)

        return event

    def __change_event_coordinate_system(self, event: Event):
        position_changes = {
            field.name: self.__change_point_coordinate_system(
                getattr(event, field.name)
            )
            for field in fields(event)
            if field.name.endswith("coordinates")
            and getattr(event, field.name)
        }

        return replace(event, **position_changes)

    def __change_event_dimensions(self, event: Event):
        position_changes = {
            field.name: self.change_point_dimensions(
                getattr(event, field.name)
            )
            for field in fields(event)
            if field.name.endswith("coordinates")
            and getattr(event, field.name)
        }

        return replace(event, **position_changes)

    def __flip_event(self, event: Event):
        position_changes = {
            field.name: self.flip_point(getattr(event, field.name))
            for field in fields(event)
            if field.name.endswith("coordinates")
            and getattr(event, field.name)
        }

        return replace(event, **position_changes)

    def get_to_coordinate_system(self) -> Optional[CoordinateSystem]:
        return self._to_coordinate_system

    @classmethod
    def transform_dataset(
        cls,
        dataset: Dataset,
        to_pitch_dimensions: Optional[PitchDimensions] = None,
        to_orientation: Optional[Orientation] = None,
        to_coordinate_system: Optional[CoordinateSystem] = None,
    ) -> Dataset:
        if (
            to_pitch_dimensions is None
            and to_orientation is None
            and to_coordinate_system is None
        ):
            return dataset

        if to_orientation is None:
            to_orientation = dataset.metadata.orientation
        elif to_orientation == Orientation.BALL_OWNING_TEAM:
            if not dataset.metadata.flags & DatasetFlag.BALL_OWNING_TEAM:
                raise ValueError(
                    "Cannot transform to BALL_OWNING_TEAM orientation when "
                    "dataset doesn't contain ball owning team data"
                )

        if to_pitch_dimensions is not None:
            # Transform the pitch dimensions and optionally the orientation
            if dataset.metadata.coordinate_system is None:
                transformer = cls(
                    from_pitch_dimensions=dataset.metadata.pitch_dimensions,
                    from_orientation=dataset.metadata.orientation,
                    to_pitch_dimensions=to_pitch_dimensions,
                    to_orientation=to_orientation,
                )
                metadata = replace(
                    dataset.metadata,
                    pitch_dimensions=to_pitch_dimensions,
                    orientation=to_orientation,
                )
            else:
                to_coordinate_system = CustomCoordinateSystem(
                    origin=dataset.metadata.coordinate_system.origin,
                    pitch_dimensions=to_pitch_dimensions,
                    vertical_orientation=dataset.metadata.coordinate_system.vertical_orientation,
                )
                transformer = cls(
                    from_coordinate_system=dataset.metadata.coordinate_system,
                    from_orientation=dataset.metadata.orientation,
                    to_coordinate_system=to_coordinate_system,
                    to_orientation=to_orientation,
                )
                metadata = replace(
                    dataset.metadata,
                    coordinate_system=to_coordinate_system,
                    orientation=to_orientation,
                )

        elif to_coordinate_system is not None:
            # Transform the coordinate system and optionally the orientation
            transformer = cls(
                from_coordinate_system=dataset.metadata.coordinate_system,
                from_orientation=dataset.metadata.orientation,
                to_coordinate_system=to_coordinate_system,
                to_orientation=to_orientation,
            )
            metadata = replace(
                dataset.metadata,
                coordinate_system=to_coordinate_system,
                orientation=to_orientation,
            )

        else:
            # Only transform the orientation
            if dataset.metadata.coordinate_system is not None:
                transformer = cls(
                    from_coordinate_system=dataset.metadata.coordinate_system,
                    from_orientation=dataset.metadata.orientation,
                    to_coordinate_system=dataset.metadata.coordinate_system,
                    to_orientation=to_orientation,
                )
            elif dataset.metadata.pitch_dimensions is not None:
                transformer = cls(
                    from_pitch_dimensions=dataset.metadata.pitch_dimensions,
                    from_orientation=dataset.metadata.orientation,
                    to_pitch_dimensions=dataset.metadata.pitch_dimensions,
                    to_orientation=to_orientation,
                )
            else:
                raise ValueError(
                    "Cannot transform orientation when the dataset doesn't "
                    "contain the pitch dimensions or a coordinate system"
                )
            metadata = replace(
                dataset.metadata,
                orientation=to_orientation,
            )

        if isinstance(dataset, TrackingDataset):
            frames = [
                transformer.transform_frame(record)
                for record in dataset.records
            ]

            return TrackingDataset(
                metadata=metadata,
                records=frames,
            )
        elif isinstance(dataset, EventDataset):
            events = [
                transformer.transform_event(event) for event in dataset.records
            ]

            return EventDataset(
                metadata=metadata,
                records=events,
            )
        else:
            raise KloppyError("Unknown Dataset type")


class DatasetTransformerBuilder:
    def __init__(
        self, to_coordinate_system: Optional[Union[str, Provider]] = None
    ):
        from kloppy.config import get_config

        if not to_coordinate_system:
            to_coordinate_system = get_config("coordinate_system")

        if not to_coordinate_system:
            to_coordinate_system = Provider.KLOPPY

        to_dataset_type = None
        if isinstance(to_coordinate_system, str):
            if ":" in to_coordinate_system:
                provider_name, dataset_type_name = to_coordinate_system.split(
                    ":"
                )
                to_coordinate_system = Provider[provider_name.upper()]
                to_dataset_type = DatasetType[dataset_type_name.upper()]
            else:
                to_coordinate_system = Provider[to_coordinate_system.upper()]

        self.to_coordinate_system = to_coordinate_system
        self.to_dataset_type = to_dataset_type

    def build(
        self,
        provider: Provider,
        dataset_type: DatasetType,
        pitch_length: Optional[float] = None,
        pitch_width: Optional[float] = None,
    ):
        from_coordinate_system = build_coordinate_system(
            # This comment forces black to keep the arguments as multi-line
            provider,
            dataset_type=dataset_type,
            pitch_length=pitch_length,
            pitch_width=pitch_width,
        )

        to_coordinate_system = build_coordinate_system(
            self.to_coordinate_system,
            dataset_type=self.to_dataset_type or dataset_type,
            pitch_length=pitch_length,
            pitch_width=pitch_width,
        )

        needs_pitch_dimensions_change = (
            from_coordinate_system.pitch_dimensions
            != to_coordinate_system.pitch_dimensions
        )
        not_standardized = (
            not from_coordinate_system.pitch_dimensions.standardized
            or not to_coordinate_system.pitch_dimensions.standardized
        )
        missing_dimensions = pitch_length is None or pitch_width is None
        if (
            needs_pitch_dimensions_change
            and not_standardized
            and missing_dimensions
        ):
            from_coordinate_system_name = (
                from_coordinate_system.provider
                if isinstance(from_coordinate_system, ProviderCoordinateSystem)
                else "custom"
            )
            to_coordinate_system_name = (
                to_coordinate_system.provider
                if isinstance(to_coordinate_system, ProviderCoordinateSystem)
                else "custom"
            )
            warnings.warn(
                "The pitch dimensions are required to transform coordinates "
                f"from {from_coordinate_system_name} to {to_coordinate_system_name}. "
                f"Using default pitch dimensions ({DEFAULT_PITCH_LENGTH} x {DEFAULT_PITCH_WIDTH}). "
                "This might result in inaccurate coordinates."
            )
            return self.build(
                provider,
                dataset_type,
                pitch_length=DEFAULT_PITCH_LENGTH,
                pitch_width=DEFAULT_PITCH_WIDTH,
            )

        return DatasetTransformer(
            from_coordinate_system=from_coordinate_system,
            to_coordinate_system=to_coordinate_system,
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from kloppy.domain.models.common import (
    AttackingDirection,
    DatasetType,
    OrientationError,
    PositionType,
)
from kloppy.domain.models.time import Time
from kloppy.utils import (
    DeprecatedEnumValue,
    camelcase_to_snakecase,
    deprecated,
    docstring_inherit_attributes,
    removes_suffix,
)

from ...exceptions import InvalidFilterError, KloppyError, OrphanedRecordError
from .common import DataRecord, Dataset, Player, Team
from .formation import FormationType
from .pitch import Point

if TYPE_CHECKING:
    from ..services.transformers.data_record import NamedColumns
    from .tracking import Frame

QualifierValueType = TypeVar("QualifierValueType")
EnumQualifierType = TypeVar("EnumQualifierType", bound=Enum)


class ResultType(Enum):
    @property
    @abstractmethod
    def is_success(self):
        raise NotImplementedError

    def __str__(self):
        return self.value


class ShotResult(ResultType):
    """
    ShotResult

    Attributes:
        GOAL (ShotResult): Shot resulted in a goal
        OFF_TARGET (ShotResult): Shot was off target
        POST (ShotResult): Shot hit the post
        BLOCKED (ShotResult): Shot was blocked by another player
        SAVED (ShotResult): Shot was saved by the keeper
        OWN_GOAL (ShotResult): Shot resulted in an own goal
    """

    GOAL = "GOAL"
    OFF_TARGET = "OFF_TARGET"
    POST = "POST"
    BLOCKED = "BLOCKED"
    SAVED = "SAVED"
    OWN_GOAL = "OWN_GOAL"

    @property
    def is_success(self):
        """
        Returns if the shot was a goal
        """
        return self == self.GOAL


class PassResult(ResultType):
    """
    PassResult

    Attributes:
        COMPLETE (PassResult): Complete pass
        INCOMPLETE (PassResult): Incomplete pass (intercepted)
        OUT (PassResult): Ball went out
        OFFSIDE (PassResult): Offside
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"
    OUT = "OUT"
    OFFSIDE = "OFFSIDE"

    @property
    def is_success(self):
        """
        Returns if the pass was complete
        """
        return self == self.COMPLETE


class TakeOnResult(ResultType):
    """
    TakeOnResult

    Attributes:
        COMPLETE (TakeOnResult): Complete take-on
        INCOMPLETE (TakeOnResult): Incomplete take-on
        OUT (TakeOnResult): Ball went out
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"
    OUT = "OUT"

    @property
    def is_success(self):
        """
        Returns if the take-on was complete
        """
        return self == self.COMPLETE


class CarryResult(ResultType):
    """
    CarryResult

    Attributes:
        COMPLETE (CarryResult): Complete carry
        INCOMPLETE (CarryResult): Incomplete carry
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"

    @property
    def is_success(self):
        """
        Returns if the carry was complete
        """
        return self == self.COMPLETE


class DuelResult(ResultType):
    """
    DuelResult

    Attributes:
        WON (DuelResult): When winning the duel (player touching the ball first)
        LOST (DuelResult): When losing the duel (opponent touches the ball first)
        NEUTRAL (DuelResult): When neither player wins duel [Mainly for WyScout v2]
    """

    WON = "WON"
    LOST = "LOST"
    NEUTRAL = "NEUTRAL"

    @property
    def is_success(self):
        """
        Returns if the duel was won
        """
        return self == self.WON


class InterceptionResult(ResultType):
    """
    InterceptionResult

    Attributes:
        SUCCESS (InterceptionResult): An interception that gains possession of the ball (without going out of bounds)
        LOST (InterceptionResult): An interception by the defending team that knocked the ball to an attacker
        OUT (InterceptionResult): An interception that knocked the ball out of bounds
    """

    SUCCESS = "SUCCESS"
    LOST = "LOST"
    OUT = "OUT"

    @property
    def is_success(self):
        """
        Returns if the interception was successful
        """
        return self == self.SUCCESS


ResultT = TypeVar("ResultT", bound=ResultType)


@dataclass
class NoResultMixin:
    """
    Mixin for event types that do not have a result attribute.
    """

    result: None

    def matches(
        self, filter_: Optional[Union[str, Callable[[Event], bool]]]
    ) -> bool:
        return super().matches(filter_)  # type: ignore


@dataclass
class ResultMixin(Generic[ResultT]):
    """
    Mixin for event types that can have a result attribute.
    """

    result: ResultT

    def matches(
        self, filter_: Optional[Union[str, Callable[[Event], bool]]]
    ) -> bool:
        if filter_ is None:
            return True
        elif callable(filter_):
            return filter_(self)  # type: ignore
        elif isinstance(filter_, str):
            """
            Allowed formats:
            1. <event_type>
            2. <event_type>.<result>

            This format always us to go to css selectors without breaking existing code.
            """
            parts = filter_.upper().split(".")
            if len(parts) == 2:
                event_type, result = parts
            elif len(parts) == 1:
                event_type = parts[0]
                result = None
            else:
                raise InvalidFilterError(
                    f"Don't know how to apply filter {filter_}"
                )

            # Call the parent `matches` for event type filtering
            if not super().matches(event_type):  # type: ignore
                return False

            if result:
                if not self.result:
                    return False

                try:
                    if self.result != self.result.__class__[result]:
                        return False
                except KeyError:
                    # result isn't applicable for this event
                    # example: result='GOAL' event=<Pass>
                    return False

            return True

    def __str__(self):
        return super().__str__()[:-1] + f" result='{self.result}'>"


class CardType(Enum):
    """
    CardType

    Attributes:
        FIRST_YELLOW (CardType): First yellow card
        SECOND_YELLOW (CardType): Second yellow card
        RED (CardType): Red card
    """

    FIRST_YELLOW = "FIRST_YELLOW"
    SECOND_YELLOW = "SECOND_YELLOW"
    RED = "RED"


class EventType(Enum):
    """
    Attributes:
        GENERIC (EventType): Unrecognised event type
        PASS (EventType):
        SHOT (EventType):
        TAKE_ON (EventType):
        CARRY (EventType):
        CLEARANCE (EventType):
        INTERCEPTION (EventType):
        DUEL (EventType):
        SUBSTITUTION (EventType):
        CARD (EventType):
        PLAYER_ON (EventType):
        PLAYER_OFF (EventType):
        RECOVERY (EventType):
        MISCONTROL (EventType):
        BALL_OUT (EventType):
        FOUL_COMMITTED (EventType):
        GOALKEEPER (EventType):
        PRESSURE (EventType):
        FORMATION_CHANGE (EventType):
    """

    GENERIC = "generic"

    PASS = "PASS"
    SHOT = "SHOT"
    TAKE_ON = "TAKE_ON"
    CARRY = "CARRY"
    CLEARANCE = "CLEARANCE"
    INTERCEPTION = "INTERCEPTION"
    DUEL = "DUEL"
    SUBSTITUTION = "SUBSTITUTION"
    CARD = "CARD"
    PLAYER_ON = "PLAYER_ON"
    PLAYER_OFF = "PLAYER_OFF"
    RECOVERY = "RECOVERY"
    MISCONTROL = "MISCONTROL"
    BALL_OUT = "BALL_OUT"
    FOUL_COMMITTED = "FOUL_COMMITTED"
    GOALKEEPER = "GOALKEEPER"
    PRESSURE = "PRESSURE"
    FORMATION_CHANGE = "FORMATION_CHANGE"

    def __repr__(self):
        return self.value


@dataclass
class Qualifier(Generic[QualifierValueType], ABC):
    """
    An event qualifier.

    Qualifiers are linked to events and provide more detailed information
    about the specific event that occurred. Each event can have a series of
    qualifiers describing it.

    Attributes:
        name (str): The name of the qualifier.
        value (object): Contains any related information.
    """

    value: QualifierValueType

    @abstractmethod
    def to_dict(self) -> Dict[str, QualifierValueType]:
        """
        Return the qualifier as a dict.
        """

    @property
    def name(self):
        return camelcase_to_snakecase(
            removes_suffix(type(self).__name__, "Qualifier")
        )


@dataclass
class BoolQualifier(Qualifier[bool], ABC):
    """
    An event qualifier with a true/false value.
    """

    def to_dict(self) -> Dict[str, bool]:
        return {f"is_{self.name}": self.value}


class EnumQualifier(Qualifier[EnumQualifierType], ABC):
    """
    An event qualifier with a set of possible values.
    """

    def to_dict(self) -> Dict[str, EnumQualifierType]:
        return {f"{self.name}_type": self.value.value}


class SetPieceType(Enum):
    """
    SetPieceType

    Attributes:
        GOAL_KICK (SetPieceType): A goal kick.
        FREE_KICK (SetPieceType): A free kick.
        THROW_IN (SetPieceType): A throw in.
        CORNER_KICK (SetPieceType): A corner kick.
        PENALTY (SetPieceType): A penalty kick.
        KICK_OFF (SetPieceType): A kick off at the beginning of a match or after scoring.
    """

    GOAL_KICK = "GOAL_KICK"
    FREE_KICK = "FREE_KICK"
    THROW_IN = "THROW_IN"
    CORNER_KICK = "CORNER_KICK"
    PENALTY = "PENALTY"
    KICK_OFF = "KICK_OFF"


@dataclass
class SetPieceQualifier(EnumQualifier[SetPieceType]):
    """
    Indicates that a pass or shot was a set piece.

    Attributes:
        name (str): `"set_piece"`
        value (SetPieceType): The type of set piece
    """


@dataclass
class CardQualifier(EnumQualifier[CardType]):
    """
    Indicates that a card was given.

    Attributes:
        name (str): `"card"`
        value (CardType): Specifies the type of card.
    """


class PassType(Enum):
    """
    PassType

    Attributes:
        CROSS (PassType): A ball played in from wide areas into the box.
        HAND_PASS (PassType): A pass given with a player’s hand.
        HEAD_PASS (PassType): A pass given with a player's head.
        HIGH_PASS (PassType): A pass that goes above shoulder level at peak height.
        LAUNCH (PassType): A long forward pass that does not appear to have a specific target.
        SIMPLE_PASS (PassType): A standard pass without complex maneuvers.
        SMART_PASS (PassType): A creative and penetrative pass attempting to break defensive lines.
        LONG_BALL (PassType): A pass that travels at least 32 meters.
        THROUGH_BALL (PassType): A pass played into space behind the defense.
        CHIPPED_PASS (PassType): A pass lifted into the air.
        FLICK_ON (PassType): A pass where a player 'flicks' the ball on towards a teammate using their head.
        ASSIST (PassType): A pass leading directly to a goal.
        ASSIST_2ND (PassType): A pass leading to another pass which then leads to a goal.
        SWITCH_OF_PLAY (PassType): Any pass which crosses the centre zone of the pitch and in length travels more than 50% of the width of the pitch.
    """

    CROSS = "CROSS"
    HAND_PASS = "HAND_PASS"
    HEAD_PASS = "HEAD_PASS"
    HIGH_PASS = "HIGH_PASS"
    LAUNCH = "LAUNCH"
    SIMPLE_PASS = "SIMPLE_PASS"
    SMART_PASS = "SMART_PASS"
    LONG_BALL = "LONG_BALL"
    THROUGH_BALL = "THROUGH_BALL"
    CHIPPED_PASS = "CHIPPED_PASS"
    FLICK_ON = "FLICK_ON"
    SHOT_ASSIST = "SHOT_ASSIST"
    ASSIST = "ASSIST"
    ASSIST_2ND = "ASSIST_2ND"
    SWITCH_OF_PLAY = "SWITCH_OF_PLAY"


@dataclass
class PassQualifier(EnumQualifier[PassType]):
    """
    Specifies the pass type.

    Attributes:
        name (str): `"pass"`
        value (PassType): The pass type.
    """


class BodyPart(Enum):
    """
    BodyPart

    Attributes:
        RIGHT_FOOT (BodyPart): Pass or Shot with right foot, save with right foot (for goalkeepers).
        LEFT_FOOT (BodyPart): Pass or Shot with left foot, save with left foot (for goalkeepers).
        HEAD (BodyPart): Pass or Shot with head, save with head (for goalkeepers).
        OTHER (BodyPart): Other body part (chest, back, etc.), for Pass and Shot.
        HEAD_OTHER (BodyPart): Pass or Shot with head or other body part. Only used when the data provider does not distinguish between HEAD and OTHER.
        BOTH_HANDS (BodyPart): Goalkeeper only. Save with both hands.
        CHEST (BodyPart): Goalkeeper only. Save with chest.
        LEFT_HAND (BodyPart): Goalkeeper only. Save with left hand.
        RIGHT_HAND (BodyPart): Goalkeeper only. Save with right hand.
        DROP_KICK (BodyPart): Pass is a keeper drop kick.
        KEEPER_ARM (BodyPart): Pass thrown from keepers hands.
        NO_TOUCH (BodyPart): Pass only. A player deliberately let the pass go past him instead of receiving it to deliver to a teammate behind him. (Also known as a "dummy").
    """

    RIGHT_FOOT = "RIGHT_FOOT"
    LEFT_FOOT = "LEFT_FOOT"
    HEAD = "HEAD"
    OTHER = "OTHER"
    HEAD_OTHER = "HEAD_OTHER"

    BOTH_HANDS = "BOTH_HANDS"
    CHEST = "CHEST"
    LEFT_HAND = "LEFT_HAND"
    RIGHT_HAND = "RIGHT_HAND"
    DROP_KICK = "DROP_KICK"
    KEEPER_ARM = "KEEPER_ARM"

    NO_TOUCH = "NO_TOUCH"


@dataclass
class BodyPartQualifier(EnumQualifier[BodyPart]):
    """
    Specifies the body part used to perform an action.

    Attributes:
        name (str): `"body_part"`
        value (BodyPart): The body part.
    """


class GoalkeeperAction(Enum):
    """
    Deprecated: GoalkeeperAction has been renamed to GoalkeeperActionType.

    Attributes:
        SAVE (GoalkeeperAction): Goalkeeper faces shot and saves.
        CLAIM (GoalkeeperAction): Goalkeeper catches cross.
        PUNCH (GoalkeeperAction): Goalkeeper punches ball clear.
        PICK_UP (GoalkeeperAction): Goalkeeper picks up ball.
        SMOTHER (GoalkeeperAction): Goalkeeper coming out to dispossess a player, equivalent to a tackle for an outfield player.
        REFLEX (GoalkeeperAction): Goalkeeper performs a reflex to save a ball.
        SAVE_ATTEMPT (GoalkeeperAction): Goalkeeper attempting to save a shot.
    """

    SAVE = DeprecatedEnumValue("SAVE")
    CLAIM = DeprecatedEnumValue("CLAIM")
    PUNCH = DeprecatedEnumValue("PUNCH")
    PICK_UP = DeprecatedEnumValue("PICK_UP")
    SMOTHER = DeprecatedEnumValue("SMOTHER")
    REFLEX = DeprecatedEnumValue("REFLEX")
    SAVE_ATTEMPT = DeprecatedEnumValue("SAVE_ATTEMPT")


class GoalkeeperActionType(Enum):
    """
    GoalkeeperActionType

    Attributes:
        SAVE (GoalkeeperActionType): Goalkeeper faces shot and saves.
        CLAIM (GoalkeeperActionType): Goalkeeper catches cross.
        PUNCH (GoalkeeperActionType): Goalkeeper punches ball clear.
        PICK_UP (GoalkeeperActionType): Goalkeeper picks up ball.
        SMOTHER (GoalkeeperActionType): Goalkeeper coming out to dispossess a player, equivalent to a tackle for an outfield player.
        REFLEX (GoalkeeperActionType): Goalkeeper performs a reflex to save a ball.
        SAVE_ATTEMPT (GoalkeeperActionType): Goalkeeper attempting to save a shot.
    """

    SAVE = "SAVE"
    CLAIM = "CLAIM"
    PUNCH = "PUNCH"
    PICK_UP = "PICK_UP"
    SMOTHER = "SMOTHER"

    REFLEX = "REFLEX"
    SAVE_ATTEMPT = "SAVE_ATTEMPT"


@dataclass
class GoalkeeperQualifier(EnumQualifier[GoalkeeperActionType]):
    """
    Specifies the goalkeeper action type.

    Attributes:
        name (str): `"goalkeeper"`
        value (GoalkeeperActionType): The goal keeper action type.
    """


class DuelType(Enum):
    """
    DuelType

    Attributes:
        AERIAL (DuelType): A duel when the ball is in the air and loose.
        GROUND (DuelType): A duel when the ball is on the ground.
        LOOSE_BALL (DuelType): When the ball is not under the control of any particular player or team.
        SLIDING_TACKLE (DuelType): A duel where the player slides on the ground to kick the ball away from an opponent.
    """

    AERIAL = "AERIAL"
    GROUND = "GROUND"
    LOOSE_BALL = "LOOSE_BALL"
    SLIDING_TACKLE = "SLIDING_TACKLE"
    TACKLE = "TACKLE"


@dataclass
class DuelQualifier(EnumQualifier[DuelType]):
    """
    Specifies the duel type.

    Attributes:
        name (str): `"duel"`
        value (DuelType): The duel type.
    """


@dataclass
class CounterAttackQualifier(BoolQualifier):
    """
    Indicates whether an event was part of a counter attack.

    Attributes:
        name (str): `"body_part"`
        value (BoolQualifier): True if the event was part of a counter attack; otherwise False.
    """


QualifierT = TypeVar("QualifierT", bound=Qualifier)


@dataclass
class NoQualifierMixin:
    """
    Mixin for event types that do not have any qualifiers.
    """

    qualifiers: None

    def get_qualifier_value(self, qualifier_type: Type[Qualifier]) -> None:
        return None

    def get_qualifier_values(self, qualifier_type: Type[Qualifier]) -> List:
        return []


@dataclass
class QualifierMixin(Generic[QualifierT]):
    """
    Mixin for event types that can have additional qualifiers.
    """

    qualifiers: List[QualifierT]

    def get_qualifier_value(self, qualifier_type: Type[QualifierT]):
        """
        Returns the Qualifier of a certain type, or None if qualifier is not present.

        Args:
            qualifier_type: one of the following QualifierTypes: [`SetPieceQualifier`][kloppy.domain.models.event.SetPieceQualifier]
                [`BodyPartQualifier`][kloppy.domain.models.event.BodyPartQualifier] [`PassQualifier`][kloppy.domain.models.event.PassQualifier]

        Examples:
            >>> from kloppy.domain import SetPieceQualifier
            >>> pass_event.get_qualifier_value(SetPieceQualifier)
            <SetPieceType.GOAL_KICK: 'GOAL_KICK'>
        """
        if self.qualifiers is None:
            return None

        for qualifier in self.qualifiers:
            if isinstance(qualifier, qualifier_type):
                return qualifier.value
        return None

    def get_qualifier_values(self, qualifier_type: Type[QualifierT]):
        """
        Returns all Qualifiers of a certain type, or None if qualifier is not present.

        Args:
            qualifier_type: one of the following QualifierTypes: [`SetPieceQualifier`][kloppy.domain.models.event.SetPieceQualifier]
                [`BodyPartQualifier`][kloppy.domain.models.event.BodyPartQualifier] [`PassQualifier`][kloppy.domain.models.event.PassQualifier]

        Examples:
            >>> from kloppy.domain import SetPieceQualifier
            >>> pass_event.get_qualifier_values(SetPieceQualifier)
            [<SetPieceType.GOAL_KICK: 'GOAL_KICK'>]
        """
        return [
            qualifier.value
            for qualifier in self.qualifiers
            if isinstance(qualifier, qualifier_type)
        ]


@dataclass
class UnderPressureQualifier(BoolQualifier):
    pass


@dataclass
@docstring_inherit_attributes(DataRecord)
class Event(DataRecord, ABC):
    """
    Abstract event baseclass. All other event classes inherit from this class.

    Attributes:
        event_id: Event identifier given by provider. Alias for `record_id`.
        event_type: The type of event.
        event_name: Name of the event type.
        time: Time in the match the event takes place.
        coordinates: Coordinates where event happened.
        team: Team related to the event.
        player: Player related to the event.
        ball_owning_team: Team in control of the ball during the event.
        ball_state: Whether the ball is in play.
        raw_event: The data provider's raw representation of the event.
        dataset: A reference to the dataset the event is part of.
        prev_record: A reference to the previous event.
        next_record: A reference to the next event.
        related_event_ids: Event identifiers of related events.
        freeze_frame: A snapshot with the location of other players at the time of the event.
        attacking_direction: The playing direction of `team` during the event.
        state: Additional game state information.
    """

    event_id: str
    team: Team
    player: Player
    coordinates: Point

    raw_event: object
    state: Dict[str, Any]
    related_event_ids: List[str]

    freeze_frame: Optional["Frame"]

    @property
    def record_id(self) -> str:
        return self.event_id

    @property
    @abstractmethod
    def event_type(self) -> EventType:
        raise NotImplementedError

    @property
    @abstractmethod
    def event_name(self) -> str:
        raise NotImplementedError

    @property
    def attacking_direction(self) -> AttackingDirection:
        if (
            self.dataset
            and self.dataset.metadata
            and self.dataset.metadata.orientation is not None
        ):
            try:
                return AttackingDirection.from_orientation(
                    self.dataset.metadata.orientation,
                    period=self.period,
                    ball_owning_team=self.ball_owning_team,
                    action_executing_team=self.team,
                )
            except OrientationError:
                return AttackingDirection.NOT_SET
        return AttackingDirection.NOT_SET

    def get_related_events(self) -> List[Event]:
        if not self.dataset:
            raise OrphanedRecordError()

        return [
            event
            for event_id in self.related_event_ids
            if (event := self.dataset.get_record_by_id(event_id)) is not None
        ]

    def get_related_event(
        self, type_: Union[str, EventType]
    ) -> Optional[Event]:
        event_type = (
            EventType[type_.upper()] if isinstance(type_, str) else type_
        )
        for related_event in self.get_related_events():
            if related_event.event_type == event_type:
                return related_event
        return None

    """Define all related events for easy access"""

    def related_pass(self) -> Optional[PassEvent]:
        return cast(
            Optional[PassEvent], self.get_related_event(EventType.PASS)
        )

    def related_shot(self) -> Optional[ShotEvent]:
        return cast(
            Optional[ShotEvent], self.get_related_event(EventType.SHOT)
        )

    def related_take_on(self) -> Optional[TakeOnEvent]:
        return cast(
            Optional[TakeOnEvent], self.get_related_event(EventType.TAKE_ON)
        )

    def related_carry(self) -> Optional[CarryEvent]:
        return cast(
            Optional[CarryEvent], self.get_related_event(EventType.CARRY)
        )

    def related_substitution(self) -> Optional[SubstitutionEvent]:
        return cast(
            Optional[SubstitutionEvent],
            self.get_related_event(EventType.SUBSTITUTION),
        )

    def related_card(self) -> Optional[CardEvent]:
        return cast(
            Optional[CardEvent], self.get_related_event(EventType.CARD)
        )

    def related_player_on(self) -> Optional[PlayerOnEvent]:
        return cast(
            Optional[PlayerOnEvent],
            self.get_related_event(EventType.PLAYER_ON),
        )

    def related_player_off(self) -> Optional[PlayerOffEvent]:
        return cast(
            Optional[PlayerOffEvent],
            self.get_related_event(EventType.PLAYER_OFF),
        )

    def related_recovery(self) -> Optional[RecoveryEvent]:
        return cast(
            Optional[RecoveryEvent], self.get_related_event(EventType.RECOVERY)
        )

    def related_ball_out(self) -> Optional[BallOutEvent]:
        return cast(
            Optional[BallOutEvent], self.get_related_event(EventType.BALL_OUT)
        )

    def related_foul_committed(self) -> Optional[FoulCommittedEvent]:
        return cast(
            Optional[FoulCommittedEvent],
            self.get_related_event(EventType.FOUL_COMMITTED),
        )

    def related_formation_change(self) -> Optional[FormationChangeEvent]:
        return cast(
            Optional[FormationChangeEvent],
            self.get_related_event(EventType.FORMATION_CHANGE),
        )

    def matches(
        self, filter_: Optional[Union[str, Callable[[Event], bool]]]
    ) -> bool:
        if filter_ is None:
            return True
        elif callable(filter_):
            return filter_(self)
        elif isinstance(filter_, str):
            """
            Allowed formats:
            1. <event_type>
            2. <event_type>.<result>

            This format always us to go to css selectors without breaking existing code.
            """
            parts = filter_.upper().split(".")
            if len(parts) == 2:
                event_type, result = parts
            elif len(parts) == 1:
                event_type = parts[0]
                result = None
            else:
                raise InvalidFilterError(
                    f"Don't know how to apply filter {filter_}"
                )

            if event_type:
                try:
                    if self.event_type != EventType[event_type]:
                        return False
                except KeyError:
                    raise InvalidFilterError(
                        f"Cannot find event type {event_type}. Possible options: {[e.value.lower() for e in EventType]}"
                    )

            if result:
                return False

            return True

    def __str__(self):
        event_type = (
            self.__class__.__name__
            if not isinstance(self, GenericEvent)
            else f"GenericEvent:{self.event_name}"
        )

        return (
            f"<{event_type} "
            f"event_id='{self.event_id}' "
            f"time='{self.time}' "
            f"team='{self.team}' "
            f"player='{self.player}'>"
        )

    def __repr__(self):
        return str(self)


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class GenericEvent(NoQualifierMixin, NoResultMixin, Event):
    """
    Unrecognised event type.

    By default, all events are deserialized as `GenericEvent` objects.
    Some events may be converted into more specialized event types, such as
    a [`PassEvent`][kloppy.domain.models.event.PassEvent] object.

    Attributes:
        event_type (EventType): `EventType.GENERIC`
        event_name (str): Defaults to `"generic"` but can be overridden with the provider's event name.
    """

    event_name: str = "generic"  # type: ignore

    @property
    def event_type(self) -> EventType:
        return EventType.GENERIC


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class ShotEvent(
    QualifierMixin[
        Union[BodyPartQualifier, SetPieceQualifier, CounterAttackQualifier]
    ],
    ResultMixin[ShotResult],
    Event,
):
    """
    An intentional attempt to score a goal by striking or directing the ball
    towards the opponent's goal. Own goals are always categorized as a shot too.

    Attributes:
        event_type (EventType): `EventType.SHOT`
        event_name (str): `"shot"`
        result (ShotResult): The outcome of the shot.
        result_coordinates (Point): End location of the shot. If a shot is blocked, this is the coordinate of the block. If the shot is saved, it is the coordinate where the keeper touched the ball.
        qualifiers: A list of qualifiers providing additional information about the shot.
    """

    result_coordinates: Optional[Point] = None

    @property
    def event_type(self) -> EventType:
        return EventType.SHOT

    @property
    def event_name(self) -> str:
        return "shot"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class PassEvent(
    QualifierMixin[
        Union[
            PassQualifier,
            BodyPartQualifier,
            SetPieceQualifier,
            CounterAttackQualifier,
        ]
    ],
    ResultMixin[PassResult],
    Event,
):
    """
    The attempted delivery of the ball from one player to another player on the same team.

    Attributes:
        event_type (EventType): `EventType.PASS`
        event_name (str): `"pass"`
        result (PassResult): The pass's outcome.
        receive_timestamp (Time): The time the pass was received.
        receiver_coordinates (Point): The coordinates where the pass was received.
        receiver_player (Player): The intended receiver of the pass.
        qualifiers: A list of qualifiers providing additional information about the pass.
    """

    receive_timestamp: Optional[Time] = None
    receiver_player: Optional[Player] = None
    receiver_coordinates: Optional[Point] = None

    @property
    def event_type(self) -> EventType:
        return EventType.PASS

    @property
    def event_name(self) -> str:
        return "pass"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class TakeOnEvent(
    QualifierMixin[CounterAttackQualifier],
    ResultMixin[TakeOnResult],
    Event,
):
    """
    An attempt by one player to dribble past an opponent.

    Attributes:
        event_type (EventType): `EventType.TAKE_ON`
        event_name (str): `"take_on"`
        result (TakeOnResult): The take-on's outcome.
        qualifiers: A list of qualifiers providing additional information about the take-on.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.TAKE_ON

    @property
    def event_name(self) -> str:
        return "take_on"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class CarryEvent(
    QualifierMixin[CounterAttackQualifier],
    ResultMixin[CarryResult],
    Event,
):
    """
    A player controls the ball at their feet while moving or standing still.

    Attributes:
        event_type (EventType): `EventType.CARRY`
        event_name (str): `"carry"`
        end_timestamp (Time): Duration of the carry.
        end_coordinates (Point): Coordinate on the pitch where the carry ended.
        result (CarryResult): The outcome of the carry.
        qualifiers: A list of qualifiers providing additional information about the carry.
    """

    end_timestamp: Time
    end_coordinates: Point

    @property
    def event_type(self) -> EventType:
        return EventType.CARRY

    @property
    def event_name(self) -> str:
        return "carry"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class InterceptionEvent(
    QualifierMixin[CounterAttackQualifier],
    ResultMixin[InterceptionResult],
    Event,
):
    """
    When a player intercepts any pass event between opposition players and
    prevents the ball reaching its target.

    Attributes:
        event_type (EventType): `EventType.INTERCEPTION`
        event_name (str): `"interception"`
        result (InterceptionResult): The outcome of the interception.
        qualifiers: A list of qualifiers providing additional information about the interception.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.INTERCEPTION

    @property
    def event_name(self) -> str:
        return "interception"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class ClearanceEvent(
    QualifierMixin[CounterAttackQualifier],
    NoResultMixin,
    Event,
):
    """
    A defensive action when a player attempts to get the ball away from
    a dangerous zone on the pitch with no immediate target regarding
    a recipient for the ball.

    Attributes:
        event_type (EventType): `EventType.CLEARANCE`
        event_name (str): `"clearance"`
        qualifiers: A list of qualifiers providing additional information about the clearance.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.CLEARANCE

    @property
    def event_name(self) -> str:
        return "clearance"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class DuelEvent(
    QualifierMixin[Union[DuelQualifier, CounterAttackQualifier]],
    ResultMixin[DuelResult],
    Event,
):
    """
    A contest between two players of opposing sides in the match. Duel events
    come in pairs: one for the attacking team, one for the defending team.
    If the duel is a take on, the offensive one has type `TAKE_ON`.

    Attributes:
        event_type (EventType): `EventType.DUEL`
        event_name (str): `"duel"`
        result (DuelResult): The outcome of the duel.
        qualifiers: A list of qualifiers providing additional information about the duel.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.DUEL

    @property
    def event_name(self) -> str:
        return "duel"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class SubstitutionEvent(NoQualifierMixin, NoResultMixin, Event):
    """
    A player is substituted off the field and replaced by another player.

    Attributes:
        event_type (EventType): `EventType.SUBSTITUTION`
        event_name (str): `"substitution"`
        replacement_player (Player): The player coming on the pitch.
    """

    replacement_player: Player
    position: Optional[PositionType] = None

    @property
    def event_type(self) -> EventType:
        return EventType.SUBSTITUTION

    @property
    def event_name(self) -> str:
        return "substitution"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class PlayerOffEvent(NoQualifierMixin, NoResultMixin, Event):
    """
    A player goes/is carried out of the pitch without a substitution.

    Attributes:
        event_type (EventType): `EventType.PLAYER_OFF`
        event_name (str): `"player_off"`
    """

    @property
    def event_type(self) -> EventType:
        return EventType.PLAYER_OFF

    @property
    def event_name(self) -> str:
        return "player_off"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class PlayerOnEvent(NoQualifierMixin, NoResultMixin, Event):
    """
    A player returns to the pitch after a `PlayerOff` event.

    Attributes:
        event_type (EventType): `EventType.PLAYER_ON`
        event_name (str): `"player_on"`
    """

    @property
    def event_type(self) -> EventType:
        return EventType.PLAYER_ON

    @property
    def event_name(self) -> str:
        return "player_on"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class CardEvent(NoQualifierMixin, NoResultMixin, Event):
    """
    When a player receives a card.

    Attributes:
        event_type (EventType): `EventType.CARD`
        event_name (str): `"card"`
        card_type (CardType): Attribute specifying the card.
    """

    card_type: CardType

    @property
    def event_type(self) -> EventType:
        return EventType.CARD

    @property
    def event_name(self) -> str:
        return "card"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class FormationChangeEvent(NoQualifierMixin, NoResultMixin, Event):
    """
    A team alters its formation

    Attributes:
        event_type (EventType): `EventType.FORMATION_CHANGE`
        event_name (str): `"card"`
        formation_type (FormationType): The new formation being used.
        player_positions (dict[Player, Position], optional): The players and their positions.
    """

    formation_type: FormationType
    player_positions: Optional[Dict[Player, PositionType]] = None

    @property
    def event_type(self) -> EventType:
        return EventType.FORMATION_CHANGE

    @property
    def event_name(self) -> str:
        return "formation_change"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class RecoveryEvent(
    QualifierMixin[CounterAttackQualifier],
    NoResultMixin,
    Event,
):
    """
    A player gathers a loose ball and gains control of possession for their team.

    Attributes:
        event_type (EventType): `EventType.RECOVERY`
        event_name (str): "recovery"
        qualifiers: A list of qualifiers providing additional information about the recovery.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.RECOVERY

    @property
    def event_name(self) -> str:
        return "recovery"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class BallOutEvent(
    QualifierMixin[CounterAttackQualifier],
    NoResultMixin,
    Event,
):
    """
    When the ball goes out of bounds.

    Attributes:
        event_type (EventType): `EventType.BALL_OUT`
        event_name (str): "ball_out"
        qualifiers: A list of qualifiers providing additional information about the ball out event.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.BALL_OUT

    @property
    def event_name(self) -> str:
        return "ball_out"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class MiscontrolEvent(
    QualifierMixin[CounterAttackQualifier],
    NoResultMixin,
    Event,
):
    """
    A player unsuccessfully controls the ball and loses possession.

    Attributes:
        event_type (EventType): `EventType.MISCONTROL`
        event_name (str): "miscontrol"
        qualifiers: A list of qualifiers providing additional information about the miscontrol.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.MISCONTROL

    @property
    def event_name(self) -> str:
        return "miscontrol"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class FoulCommittedEvent(
    QualifierMixin[Union[CounterAttackQualifier, CardQualifier]],
    NoResultMixin,
    Event,
):
    """
    Indicates a foul has been committed. The event is defined for the team that
    commits the foul.

    Attributes:
        event_type (EventType): `EventType.FOUL_COMMITTED`
        event_name (str): "foul_committed"
        qualifiers: A list of qualifiers providing additional information about the foul.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.FOUL_COMMITTED

    @property
    def event_name(self) -> str:
        return "foul_committed"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class GoalkeeperEvent(
    QualifierMixin[Union[CounterAttackQualifier, GoalkeeperQualifier]],
    NoResultMixin,
    Event,
):
    """
    Indicates an action executed by a goalkeeper.

    Attributes:
        event_type (EventType): `EventType.GOALKEEPER`
        event_name (str): "goalkeeper"
        qualifiers: A list of qualifiers providing additional information about the goalkeeper action.
    """

    @property
    def event_type(self) -> EventType:
        return EventType.GOALKEEPER

    @property
    def event_name(self) -> str:
        return "goalkeeper"


@dataclass(repr=False)
@docstring_inherit_attributes(Event)
class PressureEvent(
    QualifierMixin[CounterAttackQualifier],
    NoResultMixin,
    Event,
):
    """
    A player pressures an opponent to force a mistake.

    Attributes:
        event_type (EventType): `EventType.Pressure`
        event_name (str): `"pressure"`
        end_timestamp (Time): When the pressing ended.
        qualifiers: A list of qualifiers providing additional information about the pressure event.
    """

    end_timestamp: Time

    @property
    def event_type(self) -> EventType:
        return EventType.PRESSURE

    @property
    def event_name(self) -> str:
        return "pressure"


@dataclass(repr=False)
@docstring_inherit_attributes(Dataset)
class EventDataset(Dataset[Event]):
    """
    An event stream dataset.

    Attributes:
        dataset_type (DatasetType): `"DatasetType.EVENT"`
        events (List[Event]): A list of events. Alias for `records`.
        metadata (Metadata): Metadata for the dataset.
    """

    @property
    def dataset_type(self) -> DatasetType:
        return DatasetType.EVENT

    def _update_formations_and_positions(self):
        """Update team formations and player positions based on Substitution and TacticalShift events."""
        max_leeway = timedelta(seconds=60)

        for team in self.metadata.teams:
            team.formations.reset()

        for event in self.events:
            if isinstance(event, SubstitutionEvent):
                if event.replacement_player.starting_position:
                    replacement_player_position = (
                        event.replacement_player.starting_position
                    )
                else:
                    replacement_player_position = event.player.positions.last(
                        default=PositionType.Unknown
                    )
                event.replacement_player.set_position(
                    event.time,
                    replacement_player_position,
                )
                event.player.set_position(event.time, None)

            elif isinstance(event, FormationChangeEvent):
                if event.player_positions:
                    for player, position in event.player_positions.items():
                        if len(player.positions.items):
                            last_time, last_position = player.positions.last(
                                include_time=True
                            )
                            if last_position != position:
                                # Only update when the position changed
                                if event.time - last_time < max_leeway:
                                    # Often the formation change is detected a couple of seconds after a Substitution.
                                    # In this case we need to use the time of the Substitution
                                    player.positions.set(last_time, position)
                                else:
                                    player.positions.set(event.time, position)

                if event.team.formations.items:
                    last_time, last_formation = event.team.formations.last(
                        include_time=True
                    )
                    if last_formation != event.formation_type:
                        event.team.formations.set(
                            event.time, event.formation_type
                        )

                elif event.team.starting_formation:
                    if event.team.starting_formation != event.formation_type:
                        event.team.formations.set(
                            event.time, event.formation_type
                        )

                else:
                    event.team.formations.set(event.time, event.formation_type)

    @property
    def events(self):
        return self.records

    def get_event_by_id(self, event_id: str) -> Optional[Event]:
        return self.get_record_by_id(event_id)

    def add_state(self, *builder_keys):
        """
        See [`add_state`][kloppy.domain.services.state_builder.add_state]
        """
        from kloppy.domain.services.state_builder import add_state

        return add_state(self, *builder_keys)

    @deprecated(
        "to_pandas will be removed in the future. Please use to_df instead."
    )
    def to_pandas(
        self,
        record_converter: Optional[Callable[[Event], Dict]] = None,
        additional_columns: Optional[NamedColumns] = None,
    ) -> "DataFrame":
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "Seems like you don't have pandas installed. Please"
                " install it using: pip install pandas"
            )

        if not record_converter:
            from ..services.transformers.attribute import (
                DefaultEventTransformer,
            )

            record_converter = DefaultEventTransformer()

        def generic_record_converter(event: Event):
            row = record_converter(event)
            if additional_columns:
                for k, v in additional_columns.items():
                    if callable(v):
                        value = v(event)
                    else:
                        value = v
                    row.update({k: value})
            return row

        return pd.DataFrame.from_records(
            map(generic_record_converter, self.records)
        )

    def aggregate(self, type_: str, **aggregator_kwargs) -> List[Any]:
        if type_ == "minutes_played":
            from kloppy.domain.services.aggregators.minutes_played import (
                MinutesPlayedAggregator,
            )

            aggregator = MinutesPlayedAggregator(**aggregator_kwargs)
        else:
            raise KloppyError(f"No aggregator {type_} not found")

        return aggregator.aggregate(self)


__all__ = [
    "EnumQualifier",
    "ResultType",
    "EventType",
    "ShotResult",
    "PassResult",
    "TakeOnResult",
    "CarryResult",
    "Event",
    "GenericEvent",
    "ShotEvent",
    "PassEvent",
    "TakeOnEvent",
    "CarryEvent",
    "ClearanceEvent",
    "InterceptionEvent",
    "InterceptionResult",
    "SubstitutionEvent",
    "PlayerOnEvent",
    "PlayerOffEvent",
    "CardEvent",
    "CardType",
    "CardQualifier",
    "FormationType",
    "FormationChangeEvent",
    "EventDataset",
    "RecoveryEvent",
    "MiscontrolEvent",
    "FoulCommittedEvent",
    "BallOutEvent",
    "SetPieceType",
    "Qualifier",
    "SetPieceQualifier",
    "PassQualifier",
    "PassType",
    "BodyPart",
    "BodyPartQualifier",
    "GoalkeeperEvent",
    "GoalkeeperQualifier",
    "GoalkeeperAction",
    "GoalkeeperActionType",
    "CounterAttackQualifier",
    "UnderPressureQualifier",
    "DuelEvent",
    "DuelType",
    "DuelQualifier",
    "DuelResult",
    "BoolQualifier",
    "PressureEvent",
    "ResultMixin",
    "QualifierMixin",
]



import logging
from datetime import timedelta
import warnings
from typing import IO, NamedTuple, Optional, Union


from kloppy.domain import (
    AttackingDirection,
    BallState,
    DatasetFlag,
    Metadata,
    Orientation,
    Player,
    PlayerData,
    Point,
    Point3D,
    Provider,
    TrackingDataset,
    attacking_direction_from_frame,
)
from kloppy.domain.services.frame_factory import create_frame
from kloppy.domain import PositionType
from kloppy.exceptions import DeserializationError
from kloppy.utils import performance_logging
from kloppy.infra.serializers.event.statsperform.parsers import get_parser

from .deserializer import TrackingDataDeserializer

logger = logging.getLogger(__name__)


class StatsPerformInputs(NamedTuple):
    meta_data: IO[bytes]
    raw_data: IO[bytes]
    pitch_length: Optional[float] = None
    pitch_width: Optional[float] = None


class StatsPerformDeserializer(TrackingDataDeserializer[StatsPerformInputs]):
    def __init__(
        self,
        provider: Provider,
        limit: Optional[int] = None,
        sample_rate: Optional[float] = None,
        coordinate_system: Optional[Union[str, Provider]] = None,
        only_alive: Optional[bool] = True,
    ):
        super().__init__(limit, sample_rate, coordinate_system)
        self.only_alive = only_alive
        self._provider = provider

    @property
    def provider(self) -> Provider:
        return self._provider

    @classmethod
    def __get_frame_rate(cls, tracking):
        """Infer the frame rate of the tracking data."""

        frame_numbers = [
            int(line.split(";")[1].split(",")[0]) for line in tracking[1:]
        ]

        deltas = [
            frame_numbers[i + 1] - frame_numbers[i]
            for i in range(len(frame_numbers) - 1)
        ]

        most_common_delta = max(set(deltas), key=deltas.count)
        frame_rate = 1000 / most_common_delta

        return frame_rate

    @classmethod
    def _frame_from_framedata(cls, teams_list, period, frame_data):
        components = frame_data[1].split(":")
        frame_info = components[0].split(";")

        frame_id = int(frame_info[0])
        frame_timestamp = timedelta(
            seconds=int(frame_info[1].split(",")[0]) / 1000
        )
        match_status = int(frame_info[1].split(",")[2])

        ball_state = BallState.ALIVE if match_status == 0 else BallState.DEAD

        if len(components) > 2:
            ball_data = components[2].split(";")[0].split(",")
            ball_x, ball_y, ball_z = map(float, ball_data)
            ball_coordinates = Point3D(ball_x, ball_y, ball_z)
        else:
            ball_coordinates = None

        players_data = {}

        if components[1] != ";":  # Check if there are any players in the frame
            player_info = components[1].split(";")[:-1]
            for player_data in player_info:
                player_data = player_data.split(",")

                raw_team_side_id = int(player_data[0])
                player_id = player_data[1]
                jersey_no = int(player_data[2])
                x = float(player_data[3])
                y = float(player_data[4])

                # Field players have id 0, 1
                if raw_team_side_id in [0, 1]:
                    team_side_id = raw_team_side_id
                # Goalkeepers have id 3 and 4
                elif raw_team_side_id in [3, 4]:
                    team_side_id = raw_team_side_id - 3
                # Referees have id 2
                elif raw_team_side_id == 2:
                    continue
                else:
                    raise DeserializationError(
                        f"Unexpected team side id {raw_team_side_id}"
                    )
                team = teams_list[team_side_id]
                player = team.get_player_by_id(player_id)

                if not player:
                    player = Player(
                        player_id=player_id,
                        team=team,
                        jersey_no=jersey_no,
                        starting_position=PositionType.Unknown,
                    )
                    team.players.append(player)

                players_data[player] = PlayerData(coordinates=Point(x, y))

        frame = create_frame(
            frame_id=frame_id,
            timestamp=frame_timestamp,
            ball_coordinates=ball_coordinates,
            ball_state=ball_state,
            ball_owning_team=None,
            players_data=players_data,
            period=period,
            other_data={},
        )

        return frame

    def deserialize(self, inputs: StatsPerformInputs) -> TrackingDataset:
        with performance_logging("Loading meta data", logger=logger):
            meta_data_parser = get_parser(inputs.meta_data, "MA1")

            periods = {
                period.id: period
                for period in meta_data_parser.extract_periods()
            }
            teams_list = list(meta_data_parser.extract_lineups())
            score = meta_data_parser.extract_score()
            date = meta_data_parser.extract_date()
            game_week = meta_data_parser.extract_game_week()
            game_id = meta_data_parser.extract_game_id()

        with performance_logging("Loading tracking data", logger=logger):
            tracking_data = inputs.raw_data.read().decode("ascii").splitlines()
            frame_rate = self.__get_frame_rate(tracking_data)

            transformer = self.get_transformer(
                pitch_length=inputs.pitch_length,
                pitch_width=inputs.pitch_width,
            )

            def _iter():
                n = 0
                sample = 1.0 / self.sample_rate

                for line_ in tracking_data:
                    splits = line_.split(";")[1].split(",")
                    period_id = int(splits[1])
                    period_ = periods[period_id]
                    if n % sample == 0:
                        yield period_, line_
                    n += 1

            frames = []
            n_frames = 0
            for frame_data in _iter():
                period = frame_data[0]
                frame = self._frame_from_framedata(
                    teams_list, period, frame_data
                )
                frame = transformer.transform_frame(frame)
                if not frame.players_data or (
                    self.only_alive and frame.ball_state == BallState.DEAD
                ):
                    continue
                frames.append(frame)

                n_frames += 1

                if self.limit and n_frames >= self.limit:
                    break

        try:
            first_frame = next(
                frame for frame in frames if frame.period.id == 1
            )
            orientation = (
                Orientation.HOME_AWAY
                if attacking_direction_from_frame(first_frame)
                == AttackingDirection.LTR
                else Orientation.AWAY_HOME
            )
        except StopIteration:
            warnings.warn(
                "Could not determine orientation of dataset, defaulting to NOT_SET"
            )
            orientation = Orientation.NOT_SET

        meta_data = Metadata(
            teams=teams_list,
            periods=list(periods.values()),
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=score,
            frame_rate=frame_rate,
            orientation=orientation,
            provider=Provider.STATSPERFORM,
            flags=DatasetFlag.BALL_STATE,
            coordinate_system=transformer.get_to_coordinate_system(),
            date=date,
            game_week=game_week,
            game_id=game_id,
        )

        return TrackingDataset(
            records=frames,
            metadata=meta_data,
        )



from datetime import timedelta
from typing import Dict, List, Optional

from kloppy.domain import (
    ActionValue,
    Event,
    Frame,
    Period,
    Player,
    PlayerData,
    Point,
    Point3D,
    PositionType,
    Team,
)
from kloppy.domain.services.frame_factory import create_frame
from kloppy.exceptions import DeserializationError


def parse_str_ts(timestamp: str) -> float:
    """Parse a HH:mm:ss string timestamp into number of seconds."""
    h, m, s = timestamp.split(":")
    return timedelta(seconds=int(h) * 3600 + int(m) * 60 + float(s))


def parse_obv_values(raw_event: dict) -> Optional[ActionValue]:
    game_state_values_data = {}
    obv_mapping = {
        "obv_for_before": "action_value_scoring_before",
        "obv_against_before": "action_value_conceding_before",
        "obv_for_after": "action_value_scoring_after",
        "obv_against_after": "action_value_conceding_after",
    }
    for sb_name, kloppy_name in obv_mapping.items():
        obv_value = raw_event.get(sb_name)
        if obv_value is not None:
            game_state_values_data[kloppy_name] = obv_value

    if game_state_values_data:
        game_state_value = ActionValue(name="OBV", **game_state_values_data)

        return game_state_value


def get_team_by_id(team_id: int, teams: List[Team]) -> Team:
    """Get a team by its id."""
    if str(team_id) == teams[0].team_id:
        return teams[0]
    elif str(team_id) == teams[1].team_id:
        return teams[1]
    else:
        raise DeserializationError(f"Unknown team_id {team_id}")


def get_period_by_id(period_id: int, periods: List[Period]) -> Period:
    """Get a period by its id."""
    for period in periods:
        if period.id == period_id:
            return period
    raise DeserializationError(f"Unknown period_id {period_id}")


def parse_coordinates(
    coordinates: List[float], fidelity_version: int
) -> Point:
    """Parse coordinates into a kloppy Point.

    Coordinates are cell-based, so 1,1 (low-granularity) or 0.1,0.1
    (high-granularity) is the top-left square 'yard' of the field (in
    landscape), even though 0,0 is the true coordinate of the corner flag.

    [1, 120] x [1, 80]
    +-----+-----+
    | 1,1 | 2,1 |
    +-----+-----+
    | 1,2 | 2,2 |
    +-----+-----+
    """
    cell_side = 0.1 if fidelity_version == 2 else 1.0
    cell_relative_center = cell_side / 2
    if len(coordinates) == 2:
        return Point(
            x=coordinates[0] - cell_relative_center,
            y=coordinates[1] - cell_relative_center,
        )
    elif len(coordinates) == 3:
        # A coordinate in the goal frame, only used for the end location of
        # Shot events. The y-coordinates and z-coordinates are always detailed
        # to a tenth of a yard.
        return Point3D(
            x=coordinates[0] - cell_relative_center,
            y=coordinates[1] - 0.05,
            z=coordinates[2] - 0.05,
        )
    else:
        raise DeserializationError(
            f"Unknown coordinates format: {coordinates}"
        )


def parse_freeze_frame(
    freeze_frame: List[Dict],
    fidelity_version: int,
    home_team: Team,
    away_team: Team,
    event: Event,
    visible_area: Optional[List] = None,
) -> Frame:
    """Parse a freeze frame into a kloppy Frame."""
    players_data = {}

    def get_player_from_freeze_frame(player_data, team, i):
        if "player" in player_data:
            return team.get_player_by_id(player_data["player"]["id"])
        elif player_data.get("actor"):
            return event.player
        elif player_data.get("keeper"):
            return team.get_player_by_position(
                position=PositionType.Goalkeeper, time=event.time
            )
        else:
            return Player(
                player_id=f"T{team.team_id}-E{event.event_id}-{i}",
                team=team,
                jersey_no=None,
            )

    for i, freeze_frame_player in enumerate(freeze_frame):
        is_teammate = (event.team == home_team) == freeze_frame_player[
            "teammate"
        ]
        freeze_frame_team = home_team if is_teammate else away_team

        player = get_player_from_freeze_frame(
            freeze_frame_player, freeze_frame_team, i
        )

        players_data[player] = PlayerData(
            coordinates=parse_coordinates(
                freeze_frame_player["location"], fidelity_version
            )
        )

    if event.player not in players_data:
        players_data[event.player] = PlayerData(coordinates=event.coordinates)

    FREEZE_FRAME_FPS = 25
    frame_id = int(
        event.period.start_timestamp.total_seconds()
        + event.timestamp.total_seconds() * FREEZE_FRAME_FPS
    )

    frame = create_frame(
        frame_id=frame_id,
        ball_coordinates=Point3D(
            x=event.coordinates.x, y=event.coordinates.y, z=0
        ),
        players_data=players_data,
        period=event.period,
        timestamp=event.timestamp,
        ball_state=event.ball_state,
        ball_owning_team=event.ball_owning_team,
        other_data={"visible_area": visible_area},
    )

    return frame



from dataclasses import dataclass, field
from datetime import timedelta
from typing import Any, Callable, Dict, Optional, Union

from kloppy.domain.models.common import DatasetType
from kloppy.utils import (
    deprecated,
    docstring_inherit_attributes,
)

from .common import DataRecord, Dataset


@dataclass
@docstring_inherit_attributes(DataRecord)
class Code(DataRecord):
    """
    Single code

    Attributes:
        code_id: Unique identifier provided by the coding software. Aias for `record_id`.
        code: A string describing the code.
        end_timestamp: End timestamp for the period of time this code instance is active.
        labels: Text labels describing this code instance.
    """

    code_id: str
    code: str
    end_timestamp: timedelta
    labels: Dict[str, Union[bool, str]] = field(default_factory=dict)

    @property
    def record_id(self) -> str:
        return self.record_id

    @property
    def start_timestamp(self):
        return self.timestamp


@dataclass
class CodeDataset(Dataset[Code]):
    """
    A dataset containing SportsCode annotations.

    Attributes:
        dataset_type (DatasetType): `"DatasetType.CODE"`
        codes (List[Code]): A list of codes. Alias for `records`.
        metadata (Metadata): Metadata of the code dataset.
    """

    dataset_type: DatasetType = DatasetType.CODE

    @property
    def codes(self):
        return self.records

    @deprecated(
        "to_pandas will be removed in the future. Please use to_df instead."
    )
    def to_pandas(
        self,
        record_converter: Optional[Callable[[Code], Dict]] = None,
        additional_columns=None,
    ) -> "DataFrame":
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "Seems like you don't have pandas installed. Please"
                " install it using: pip install pandas"
            )

        if not record_converter:
            from ..services.transformers.attribute import (
                DefaultCodeTransformer,
            )

            record_converter = DefaultCodeTransformer()

        def generic_record_converter(code: Code):
            row = record_converter(code)
            if additional_columns:
                for k, v in additional_columns.items():
                    if callable(v):
                        value = v(code)
                    else:
                        value = v
                    row.update({k: value})

            return row

        return pd.DataFrame.from_records(
            map(generic_record_converter, self.records)
        )


__all__ = ["Code", "CodeDataset"]



"""XML parser for Opta F7 feeds."""

from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from lxml import objectify

from kloppy.domain import (
    FormationType,
    Ground,
    Period,
    Player,
    Score,
    Team,
    PositionType,
)
from kloppy.exceptions import DeserializationError

from .base import OptaXMLParser
from ..formation_mapping import (
    formation_position_mapping,
    formation_name_mapping,
    FormationType,
)

document_path = objectify.ObjectPath("SoccerFeed.SoccerDocument")
matchdata_path = objectify.ObjectPath("SoccerFeed.SoccerDocument.MatchData")


def _parse_f7_datetime(dt_str: str) -> datetime:
    return datetime.strptime(dt_str, "%Y%m%dT%H%M%S%z")


class F7XMLParser(OptaXMLParser):
    """Extract data from a Opta F7 data stream.

    Parameters
    ----------
    path : str
        Path of the data file.
    """

    def extract_periods(self) -> List[Period]:
        periods = {
            i: Period(id=i, start_timestamp=None, end_timestamp=None)
            for i in range(1, 5)
        }
        match_stats = matchdata_path.find(self.root).iterchildren("Stat")
        for stat in match_stats:
            if stat.attrib["Type"] == "first_half_start":
                periods[1].start_timestamp = _parse_f7_datetime(stat.text)
            elif stat.attrib["Type"] == "first_half_end":
                periods[1].end_timestamp = _parse_f7_datetime(stat.text)
            elif stat.attrib["Type"] == "second_half_start":
                periods[2].start_timestamp = _parse_f7_datetime(stat.text)
            elif stat.attrib["Type"] == "second_half_end":
                periods[2].end_timestamp = _parse_f7_datetime(stat.text)
            elif stat.attrib["Type"] == "first_half_extra_start":
                periods[3].start_timestamp = _parse_f7_datetime(stat.text)
            elif stat.attrib["Type"] == "first_half_extra_end":
                periods[3].end_timestamp = _parse_f7_datetime(stat.text)
            elif stat.attrib["Type"] == "second_half_extra_start":
                periods[4].start_timestamp = _parse_f7_datetime(stat.text)
            elif stat.attrib["Type"] == "second_half_extra_end":
                periods[4].end_timestamp = _parse_f7_datetime(stat.text)
        periods = [
            period
            for period in periods.values()
            if period.start_timestamp is not None
        ]

        result_elements = self.root.xpath(
            "/SoccerFeed/SoccerDocument/MatchData/MatchInfo/Result"
        )
        if result_elements and "Type" in result_elements[0].attrib:
            match_result_type = result_elements[0].attrib["Type"]
        else:
            match_result_type = None
        if match_result_type == "PenaltyShootout":
            periods.append(
                Period(id=5, start_timestamp=None, end_timestamp=None)
            )

        return periods

    def extract_score(self) -> Optional[Score]:
        team_elms = matchdata_path.find(self.root).iterchildren("TeamData")
        home_score = None
        away_score = None
        for team_elm in team_elms:
            if team_elm.attrib["Side"] == "Home":
                home_score = int(team_elm.attrib["Score"])
            elif team_elm.attrib["Side"] == "Away":
                away_score = int(team_elm.attrib["Score"])
            else:
                raise DeserializationError(
                    f"Unknown side: {team_elm.attrib['Side']}"
                )
        if home_score is None or away_score is None:
            return None
        return Score(home=home_score, away=away_score)

    def extract_lineups(self) -> Tuple[Team, Team]:
        """Return a dictionary with all available teams.

        Returns
        -------
        dict
            A mapping between team IDs and the information available about
            each team in the data stream.
        """
        team_elms = matchdata_path.find(self.root).iterchildren("TeamData")
        home_team = None
        away_team = None
        for team_elm in team_elms:
            if team_elm.attrib["Side"] == "Home":
                home_team = self._team_from_xml_elm(team_elm)
            elif team_elm.attrib["Side"] == "Away":
                away_team = self._team_from_xml_elm(team_elm)
            else:
                raise DeserializationError(
                    f"Unknown side: {team_elm.attrib['Side']}"
                )
        if home_team is None:
            raise DeserializationError("Could not find home team")
        if away_team is None:
            raise DeserializationError("Could not find away team")
        if len(home_team.players) == 0 or len(away_team.players) == 0:
            raise DeserializationError("Lineup incomplete")
        return home_team, away_team

    def _team_from_xml_elm(self, team_elm: Any) -> Team:
        # This should not happen here
        team_name, team_players = self._parse_team_players(
            team_elm.attrib["TeamRef"]
        )
        team = Team(
            team_id=str(team_elm.attrib["TeamRef"].lstrip("t")),
            name=team_name,
            ground=(
                Ground.HOME
                if team_elm.attrib["Side"] == "Home"
                else Ground.AWAY
            ),
            starting_formation=FormationType(
                "-".join(list(team_elm.attrib["Formation"]))
            ),
        )
        players = []

        formation = (
            formation_name_mapping.get(
                team_elm.attrib["Formation"], FormationType.UNKNOWN
            )
            if "Formation" in team_elm.attrib
            else FormationType.UNKNOWN
        )

        for player_elm in team_elm.find("PlayerLineUp").iterchildren(
            "MatchPlayer"
        ):
            starting_position = (
                formation_position_mapping[formation][
                    int(player_elm.attrib["Formation_Place"])
                ]
                if formation != FormationType.UNKNOWN
                else PositionType.Unknown
            )
            players.append(
                Player(
                    player_id=player_elm.attrib["PlayerRef"].lstrip("p"),
                    team=team,
                    jersey_no=int(player_elm.attrib["ShirtNumber"]),
                    first_name=team_players[player_elm.attrib["PlayerRef"]][
                        "first_name"
                    ],
                    last_name=team_players[player_elm.attrib["PlayerRef"]][
                        "last_name"
                    ],
                    starting=(player_elm.attrib["Status"] == "Start"),
                    starting_position=starting_position,
                )
            )

        team.players = players
        return team

    def _parse_team_players(
        self, team_ref: str
    ) -> Tuple[str, Dict[str, Dict[str, str]]]:
        team_elms = list(document_path.find(self.root).iterchildren("Team"))
        for team_elm in team_elms:
            if team_elm.attrib["uID"] == team_ref:
                team_name = str(team_elm.find("Name"))
                players = {
                    player_elm.attrib["uID"]: dict(
                        first_name=str(
                            player_elm.find("PersonName").find("First")
                        ),
                        last_name=str(
                            player_elm.find("PersonName").find("Last")
                        ),
                    )
                    for player_elm in team_elm.iterchildren("Player")
                }
                break
        else:
            raise DeserializationError(
                f"Could not parse players for {team_ref}"
            )

        return team_name, players



from abc import ABC, abstractmethod
from typing import Optional, List, Generic, TypeVar, Union

from kloppy.domain import (
    EventDataset,
    Event,
    EventType,
    DatasetTransformer,
    Provider,
    EventFactory,
    DatasetType,
    DatasetTransformerBuilder,
)

T = TypeVar("T")


class EventDataDeserializer(ABC, Generic[T]):
    def __init__(
        self,
        event_types: Optional[List[Union[EventType, str]]] = None,
        coordinate_system: Optional[Union[str, Provider]] = None,
        event_factory: Optional[EventFactory] = None,
    ):
        if not event_types:
            event_types = []

        self.event_types = [
            (
                EventType[event_type.upper()]
                if isinstance(event_type, str)
                else event_type
            )
            for event_type in event_types
        ]

        self.transformer_builder = DatasetTransformerBuilder(coordinate_system)

        if not event_factory:
            event_factory = EventFactory()
        self.event_factory = event_factory

    def should_include_event(self, event: Event) -> bool:
        if not self.event_types:
            return True
        return event.event_type in self.event_types

    def get_transformer(
        self,
        pitch_length: Optional[float] = None,
        pitch_width: Optional[float] = None,
        provider: Optional[Provider] = None,
    ) -> DatasetTransformer:
        return self.transformer_builder.build(
            provider=provider or self.provider,
            dataset_type=DatasetType.EVENT,
            pitch_length=pitch_length,
            pitch_width=pitch_width,
        )

    @property
    @abstractmethod
    def provider(self) -> Provider:
        raise NotImplementedError

    @abstractmethod
    def deserialize(self, inputs: T) -> EventDataset:
        raise NotImplementedError



from .deserializer_v2 import WyscoutDeserializerV2
from .deserializer_v3 import WyscoutDeserializerV3, WyscoutInputs

__all__ = [
    "WyscoutDeserializerV2",
    "WyscoutDeserializerV3",
    "WyscoutInputs",
]



import fsspec

from .fsspec import FSSpecAdapter


class FileAdapter(FSSpecAdapter):
    def supports(self, url: str) -> bool:
        return self._infer_protocol(url) == "file"

    def _get_filesystem(
        self, url: str, no_cache: bool = False
    ) -> fsspec.AbstractFileSystem:
        return fsspec.filesystem("file")



from abc import ABC, abstractmethod
from typing import BinaryIO, List


class Adapter(ABC):
    @abstractmethod
    def supports(self, url: str) -> bool:
        pass

    @abstractmethod
    def is_directory(self, url: str) -> bool:
        pass

    @abstractmethod
    def is_file(self, url: str) -> bool:
        pass

    @abstractmethod
    def read_to_stream(self, url: str, output: BinaryIO):
        pass

    @abstractmethod
    def list_directory(self, url: str, recursive: bool = True) -> List[str]:
        pass


__all__ = ["Adapter"]



from datetime import timedelta

import pytest

from kloppy.domain import (
    Provider,
    Orientation,
    Point,
    DatasetType,
)

from kloppy import metrica


class TestMetricaCsvTracking:
    """"""

    @pytest.fixture
    def home_data(self, base_dir) -> str:
        return base_dir / "files/metrica_home.csv"

    @pytest.fixture
    def away_data(self, base_dir) -> str:
        return base_dir / "files/metrica_away.csv"

    def test_correct_deserialization(self, home_data: str, away_data: str):
        dataset = metrica.load_tracking_csv(
            home_data=home_data, away_data=away_data
        )
        assert dataset.metadata.provider == Provider.METRICA
        assert dataset.dataset_type == DatasetType.TRACKING
        assert len(dataset.records) == 6
        assert len(dataset.metadata.periods) == 2
        assert dataset.metadata.orientation == Orientation.HOME_AWAY
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == timedelta(
            seconds=0.0
        )
        assert dataset.metadata.periods[0].end_timestamp == timedelta(
            seconds=0.12
        )
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == timedelta(
            seconds=5800.12
        )
        assert dataset.metadata.periods[1].end_timestamp == timedelta(
            seconds=5800.24
        )

        # check timestamps
        assert dataset.records[0].frame_id == 1  # period 1
        assert dataset.records[0].timestamp == timedelta(seconds=0.04)
        assert dataset.records[1].timestamp == timedelta(seconds=0.08)
        assert dataset.records[3].frame_id == 145004  # period 2
        assert dataset.records[3].timestamp == timedelta(seconds=0.04)

        # make sure data is loaded correctly (including flip y-axis)
        home_player = dataset.metadata.teams[0].players[0]
        assert dataset.records[0].players_data[
            home_player
        ].coordinates == Point(x=0.00082, y=1 - 0.48238)

        away_player = dataset.metadata.teams[1].players[0]
        assert dataset.records[0].players_data[
            away_player
        ].coordinates == Point(x=0.90509, y=1 - 0.47462)

        assert dataset.records[0].ball_coordinates == Point(
            x=0.45472, y=1 - 0.38709
        )

        # make sure player data is only in the frame when the player is at the pitch
        assert "home_14" not in [
            player.player_id
            for player in dataset.records[0].players_data.keys()
        ]
        assert "home_14" in [
            player.player_id
            for player in dataset.records[3].players_data.keys()
        ]
